import logging
from typing import List, Optional, Dict, Type


from pycognaize.common.enums import (
    IqDocumentKeysEnum,
    IqTagKeyEnum,
    ID,
    IqFieldKeyEnum,
    IqDataTypesEnum
)
from pycognaize.document.html_info import HTML
from pycognaize.document.page import Page
from pycognaize.document.field import Field
from pycognaize.document.tag import ExtractionTag
from pycognaize.document.tag.html_tag import TDTag


class NumericField(Field):
    """Base class for all pycognaize number fields"""
    tag_class: Type[ExtractionTag] = ExtractionTag
    html_tag_class: Type[TDTag] = TDTag

    def __init__(self,
                 name: str,
                 value: str = '',
                 tags: Optional[List[ExtractionTag]] = None,
                 field_id: Optional[str] = None,
                 group_key: str = None,
                 confidence: Optional[float] = -1.0,
                 group_name: str = None
                 ):
        super().__init__(name=name, tags=tags, value=value,
                         group_key=group_key, confidence=confidence,
                         group_name=group_name)
        self._field_id = field_id
        self._value = self.convert_to_numeric(value)
        if self.tags:
            self._value = sum([self.convert_to_numeric(i.raw_value)
                               for i in self.tags])

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value

    @classmethod
    def construct_from_raw(cls, raw: dict, pages: Dict[int, Page],
                           html: Optional[HTML] = None) -> 'NumericField':
        """Create NumericField object from dictionary"""
        tag_dicts: List[dict] = raw[IqDocumentKeysEnum.tags.value]
        tags = []
        for i in tag_dicts:
            try:
                if pages:
                    tags.append(cls.tag_class.construct_from_raw(
                        raw=i, page=pages[i['page']]))
                else:
                    tags.append(cls.html_tag_class.construct_from_raw(
                        raw=i, html=html))
            except Exception as e:
                logging.debug(f"Failed creating tag for field {raw[ID]}: {e}")
        value = (tags[0].raw_value if (html.path and tags)
                 else raw[IqTagKeyEnum.value.value])
        return cls(name=raw[IqDocumentKeysEnum.name.value],
                   value=value,
                   tags=tags,
                   field_id=str(raw[ID]),
                   group_key=raw.get(IqFieldKeyEnum.group_key.value, ''),
                   group_name=raw.get(IqFieldKeyEnum.group.value, '')
                   )

    @staticmethod
    def convert_to_numeric(value):
        """converts string value to numeric"""
        # noinspection PyBroadException
        try:
            value = float(value)
        except Exception:
            value = float('nan')
        return value

    def to_dict(self) -> dict:
        """Converts NumericField object to dictionary"""
        field_dict = super().to_dict()
        field_dict[ID] = self._field_id
        field_dict[IqFieldKeyEnum.value.value] = self.value
        field_dict[
            IqFieldKeyEnum.data_type.value] = IqDataTypesEnum.number.value
        return field_dict

    def __repr__(self):
        return (f"<{self.__class__.__name__}: {self.name}:"
                f" {'|'.join([i.raw_value for i in self.tags])}>")

    def __str__(self):
        return f"{'|'.join([i.raw_value for i in self.tags])}"



import enum

from .field import Field
from .date_field import DateField
from .numeric_field import NumericField
from .section_field import SectionField
from .table_field import TableField
from .text_field import TextField
from .area_field import AreaField


__all__ = ['Field', 'DateField', 'NumericField',
           'TableField', 'TextField', 'AreaField', 'FieldMapping']


class FieldMapping(enum.Enum):
    """Mapping of IQ field type names and python classes"""
    number = NumericField
    date = DateField
    text = TextField
    table = TableField
    area = AreaField
    section = SectionField



class DrawOnShellException(Exception):
    pass


class ServerAPIException(Exception):
    def __init__(self, failure_reason: str = ''):
        message = f"Connection failed due to {failure_reason}"
        super().__init__(message)


class AuthenthicationError(Exception):
    def __init__(self):
        message = "Please login to access a remote Snapshot"
        super().__init__(message)



import argparse

from src.model import CurrentModel as GenieModel

parser = argparse.ArgumentParser(description='Execute and'
                                 ' evaluate genie models')
parser.add_argument('action', metavar='action', type=str,
                    help='an action to be run on the task'
                         ' [execute | evaluate]')
parser.add_argument('--task_id', type=str, required=True, nargs='*')
parser.add_argument('--base_doc_task_id', type=str)
parser.add_argument('--token', type=str, required=True)
parser.add_argument('--url', type=str, required=True)
parser.add_argument('--model_version', type=str)

parser.add_argument('-bulk', '--bulk',  action='store_true')
parser.add_argument('-based_on_match', '--based_on_match', action='store_true')


def main():
    """Argparser for pycognaize"""
    args = parser.parse_args()

    action = args.action
    task_ids, token, url = args.task_id, args.token, args.url
    base_doc_task_id = args.base_doc_task_id
    print(base_doc_task_id)

    if action == 'execute':
        run_model(task_ids, token, url, base_doc_task_id, args.based_on_match)
    elif action == 'evaluate':
        model_version = args.base_doc_task_id
        evaluate_model(model_version, token, url)
    else:
        print('Unknown action')


def run_model(task_ids, token, url, base_doc_task_id, based_on_match):
    """Execute model given parameters from cli"""
    for task_id in task_ids:
        if based_on_match:
            if base_doc_task_id is None:
                raise ValueError('base_doc_task_id is required when'
                                 ' based_on_match is True')
            GenieModel().execute_based_on_match(task_id,
                                                base_doc_task_id,
                                                token, url)
        else:
            GenieModel().execute_genie_v2(task_id, token, url)


def evaluate_model(token, url, model_version):
    """Run Model evaluation given parameters from cli"""
    if model_version is None:
        raise ValueError('model_version is required when action is evaluate')
    GenieModel().evaluate(token, url, model_version)



"""This package is a powerful python API for creating ML/DL models using
    all the data and functionality provided by the pycognaize application.
"""

__version__ = "1.3.3"

__all__ = ['Login', 'Model', 'Snapshot']

import os
from .model import Model
from .login import Login
from .document import Snapshot

name = os.path.basename(os.path.dirname(__file__))



import bson
import string

import pandas as pd
from typing import Optional, Tuple

from pycognaize.common.enums import (
    IqCellKeyEnum,
    IqTableTagEnum,
    IqTagKeyEnum,
    ID
)
from pycognaize.document.tag import ExtractionTag
from pycognaize.document.tag.tag import BoxTag
from pycognaize.document.tag.cell import Cell
from pycognaize.common.utils import convert_coord_to_num
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pycognaize.document.page import Page


class TableTag(BoxTag):
    """Base class for all pycognaize table fields"""

    def __init__(self, left, right, top, bottom,
                 page: 'Page',
                 cell_data: dict):
        super().__init__(left=left, right=right, top=top, bottom=bottom,
                         page=page)
        self._cell_data = cell_data
        self._cells = {}
        self._populate_cells()
        self._build_df()
        self._raw_df = None
        self._df = None

    def __getitem__(self, val: Tuple[int, int]) -> Cell:
        """Gets the cell through index"""
        if len(val) == 2:
            if any([isinstance(i, slice) for i in val]):
                raise NotImplementedError("Slice lookup not implemented")
            elif val in self.cells.keys():
                return self.cells[val]
            raise IndexError(
                f"No cell with the following index in the table: {val}")
        else:
            raise ValueError(f"Invalid argument {val}")

    @property
    def cell_data(self) -> dict:
        if not self._cell_data:
            raise Exception('Cell data is empty')
        return self._cell_data

    @property
    def cells(self) -> dict:
        return self._cells

    @property
    def raw_df(self) -> pd.DataFrame:
        if self._raw_df is None:
            self._raw_df = self._build_df()
        return self._raw_df

    @property
    def df(self) -> pd.DataFrame:
        if self._df is None:
            self._df = self.raw_df.applymap(lambda x: self._extract_raw_ocr(x))
        return self._df

    @staticmethod
    def _extract_raw_ocr(x):
        """Returns OCR extraction data"""
        try:
            return x.raw_ocr_value
        except AttributeError:
            return ''

    @classmethod
    def construct_from_raw(cls, raw: dict, page: 'Page') -> 'TableTag':
        """Build Field from pycognaize raw data

        :param raw: pycognaize field's tag info
        :param page: `Page` to which the tag belongs
        :return:
        """
        table_raw_data = raw[IqTableTagEnum.table.value]
        left = convert_coord_to_num(table_raw_data['left'])
        top = convert_coord_to_num(table_raw_data['top'])
        height = convert_coord_to_num(table_raw_data['height'])
        width = convert_coord_to_num(table_raw_data['width'])
        cells = table_raw_data[IqTableTagEnum.cells.value]
        right = left + width
        bottom = top + height
        return cls(left=left, right=right, top=top, bottom=bottom,
                   page=page,
                   cell_data=cells)

    def to_dict(self) -> dict:
        """Converts table tag to dict"""
        table_dict = {
            IqTableTagEnum.page.value: self.page.page_number,
            IqTableTagEnum.left.value: f"{self.left}%",
            IqTableTagEnum.top.value: f"{self.top}%",
            IqTableTagEnum.height.value: f"{self.height}%",
            IqTableTagEnum.width.value: f"{self.width}%",
            IqTableTagEnum.cells.value: self.cell_data,
        }
        output_dict = {
            ID: bson.ObjectId(),
            IqTagKeyEnum.page.value: self.page.page_number,
            IqTagKeyEnum.ocr_value.value:
                f"table on page {self.page.page_number}",
            IqTagKeyEnum.value.value:
                f"table on page {self.page.page_number}",
            IqTagKeyEnum.is_table.value: True,
            IqTableTagEnum.table.value: table_dict
        }
        return output_dict

    def _populate_cells(self):
        for left_col_top_row, cell_dict in self.cell_data.items():
            keys = tuple((int(i) for i in left_col_top_row.split(':')))
            self._cells[keys] = (
                self._populate_cell(keys=keys, cell_dict=cell_dict))

    @staticmethod
    def _populate_cell(keys: tuple, cell_dict: dict) -> Cell:
        """Creates Cell object for each item in Table"""
        for key in IqCellKeyEnum:
            if key == IqCellKeyEnum.left_col_top_row:
                continue
            if key.value not in cell_dict:
                raise KeyError(
                    f"Required key '{key.value}' not in cell: {cell_dict}")

        return Cell(
            value=cell_dict[IqCellKeyEnum.text.value],
            left_col=keys[0],
            top_row=keys[1],
            top=cell_dict[IqCellKeyEnum.top.value],
            right=cell_dict[IqCellKeyEnum.left.value] + cell_dict[
                IqCellKeyEnum.width.value],
            bottom=cell_dict[IqCellKeyEnum.top.value] + cell_dict[
                IqCellKeyEnum.height.value],
            left=cell_dict[IqCellKeyEnum.left.value],
            col_span=cell_dict[IqCellKeyEnum.col_span.value],
            row_span=cell_dict[IqCellKeyEnum.row_span.value]
        )

    def _build_df(self, use_ocr_text: bool = False) -> pd.DataFrame:
        """Build pandas data frame using `TableTag` Cells

        :param use_ocr_text: If true, the raw OCR data will be used for
            the content of the cells
        :return: DataFrame object,where each cell contains an ExtractionTag
            object with the coordinates and values from the annotated
            document
        """
        cols = set()
        rows = set()
        image_width = self.page.image_width
        image_height = self.page.image_height
        region_in_pixels = dict(page=self.page)
        if image_width > image_height:
            image_width, image_height = image_height, image_width

        for cell_ in self.cells.values():
            cols.add(cell_.left)
            rows.add(cell_.top)

        cols = list(cols)
        rows = list(rows)
        cols.sort()
        rows.sort()

        headers_df = list(range(len(cols)))
        indices_df = list(range(len(rows)))
        df = pd.DataFrame(columns=headers_df, index=indices_df)

        for cell_ in self.cells.values():
            text = cell_.value
            top_index = rows.index(cell_.top)
            left_index = cols.index(cell_.left)
            for col_n in range(left_index, left_index + cell_.col_span):
                for row_n in range(top_index, top_index + cell_.row_span):
                    if use_ocr_text:
                        x = cell_.left * image_width / 100
                        y = cell_.top * image_height / 100,
                        w = (cell_.right - cell_.left) * image_width / 100,
                        h = (cell_.bottom - cell_.top) * image_height / 100,
                        region_in_pixels.update(
                            dict(x=x, y=y, w=w, h=h, width_scale=image_width,
                                 height_scale=image_height))
                        # FIXME: Define get_ocr_for_region and use it here
                        text = self.page.get_ocr_for_cell(region_in_pixels)
                    if not pd.isnull(df.iloc[row_n, col_n]):
                        raise ValueError(
                            "table_tag provides multiple values"
                            " for the same cell.")
                    df[col_n][row_n] = df[col_n][row_n] = ExtractionTag(
                        left=cell_.left, right=cell_.right,
                        top=cell_.top, bottom=cell_.bottom,
                        page=self.page,
                        raw_value=cell_.value,
                        raw_ocr_value=text)
        return df

    @staticmethod
    def _is_ascii(str_) -> bool:
        return not any((i for i in str_ if i not in string.ascii_letters))

    def letter_2_num(self, letters) -> int:
        """Convert excel style coordinates into zero index coordinate"""
        letters = letters.upper()
        res = 0
        if self._is_ascii(letters):
            weight = len(letters) - 1
            for i, c in enumerate(letters):
                res += (ord(c) - 64) * 26 ** (weight - i)
        return res

    @staticmethod
    def split_excel_letters_numbers(
            str_coord: str) -> Optional[Tuple[str, int]]:
        """Return the letters and numbers of the Excel coordinate as a tuple.
            If the string is not a valid Excel coordinate, return None"""
        for n, ch in enumerate(str_coord):
            if ch.isdigit():
                letters, numbers = str_coord[:n], str_coord[n:]
                if letters.isalpha() and numbers.isdigit():
                    return letters, int(numbers)
                else:
                    return



"""Defines NumericParser for parsing a number string into a float values"""
import re
from typing import Union


class NumericParser:
    """
    Parser for any kind of text input into a corresponding format.
    Numeric data is assumed to be valid and clean.
    Use NumericColumnFormatter for validating your data.
    """
    numeric = (r"(^0+$|^[\\[\\(+-]*(([1-9][0-9]{0,2}([.,][0-9]{3})*)|[0-9]"
               r"|[1-9][0-9]+)([,.][0-9]*)?[\]\)]*$)")
    delimiters = r"[.,]"
    REGEX_NO_ALPHANUM_CHARS = re.compile(r'[^a-zA-Z0-9)\[\](-.,]')

    brackets = r"[\[\]\(\)]"
    # match a pattern with whitespace followed by either a single comma
    #   from the left or whitespace surrounded by numbers
    single_white_space_regex = "(?<=\\d)(,{0,1}) {1,2}(?=\\d)"

    def __init__(self, raw: str):
        """Takes string as an input and create ValueParser object
        :param raw: input raw string representation of the number

        """

        raw = re.sub(self.single_white_space_regex, '', raw)
        raw = raw.replace('$', '').strip()
        self.raw = raw
        self.format = None
        self.parsed = None
        self.value_type = None
        self._separators = None
        self._separated_digits = None
        self.sign = None
        self.removed_sign = False

    def __get_separated_digits_and_separators(self):
        """Assigns the list of separators and separated digits to self.
            Ex. self.raw=12,345.6
            [',','.']->self.separators
            ['12', '345', '6']->self._separated_digits
            Applicable only if type is numeric
        """
        # TODO: check if strip_value is used properly
        val = self.REGEX_NO_ALPHANUM_CHARS.sub('', self.raw)
        val = re.sub(self.brackets, '', val)
        val = self.strip_value(val)
        # strip from multiple zeros. ex: 000 -> 0
        val = val[:-1].lstrip('0') + val[-1]
        self._separators = re.findall(self.delimiters, val)
        self._separated_digits = re.split(self.delimiters, val)

    def infer_sign(self):
        """
        Infer the sign of the number either with brackets
        or a trailing minus sign"""
        self.sign = 1
        if self.raw.startswith('(') and self.raw.endswith(')'):
            self.sign = -1
            self.remove_sign_parentheses()

        elif self.raw.lstrip().startswith('-'):
            self.sign = -1
            self.remove_dash_sign()

    def remove_dash_sign(self):
        self.raw = self.raw.lstrip().lstrip('-')
        self.removed_sign = True

    def remove_sign_parentheses(self):
        self.raw = self.raw.lstrip('(')
        self.raw = self.raw.rstrip(')')
        self.removed_sign = True

    def parse_regular_float(self):
        """Parse Like a regular float number"""
        # noinspection PyBroadException
        try:
            # try parsing like a regular float number
            # if it doesn't work use more complicated logic
            delimiter = any(re.findall(self.delimiters, self.raw))
            decimal_part = len(re.split(self.delimiters, self.raw)[-1]) if\
                delimiter else 0
            if decimal_part < 3:
                self.parsed = float(self.raw.strip()) * self.sign
            elif decimal_part > 2:
                return None
            return self.parsed
        except Exception:
            return None

    def parse_regular_float_with_semicolon(self):
        """Parse Like a regular float number with semicolon"""
        # noinspection PyBroadException
        try:
            if self.raw.strip().startswith('0'):
                # try parsing like a regular float number after replacing
                # commas if it doesn't work use more complicated logic
                if self.removed_sign:
                    self.parsed = float(self.raw.strip().replace(',', '.')) \
                                  * self.sign
                else:
                    self.parsed = float(self.raw.strip().replace(',', '.'))

                return self.parsed
        except Exception:
            return None

    def parse_numeric(self) -> Union[int, float]:
        """
        Parse the raw data, and return parsed numeric value
        and get the numeric, separator and value type of the raw data

        :return: Parsed float or int value
        """
        self.raw = self.raw.replace('–', '-')
        self.infer_sign()

        parse_float = self.parse_regular_float()
        if parse_float is not None:
            return parse_float
        else:
            parse_float_semicolon = self.parse_regular_float_with_semicolon()
            if parse_float_semicolon is not None:
                return parse_float_semicolon

        if '%' in self.raw:
            return float('nan')

        try:
            self.__get_separated_digits_and_separators()

            self.parsed = self.parse_raw_numeric()
            self.parsed = self.parsed * self.sign
        except Exception:
            return float('nan')
        return self.parsed

    def parse_raw_numeric(self) -> Union[int, float]:
        """Used for parsing raw text without getting any context
            from the column
        """
        if len(self._separators) == 0:
            return int(self._separated_digits[0])
        elif len(self._separated_digits[-1]) < 3:
            int_part = int(''.join(self._separated_digits[:-1]))
            dec_part = float('0.' + self._separated_digits[-1])
            return int_part + dec_part
        elif len(self._separators) > 1 and self._separators[0] != \
                self._separators[-1] and len(self._separated_digits[-1]) < 3:
            int_part = int(''.join(self._separated_digits[:-1]))
            dec_part = float('0.' + self._separated_digits[-1])
            return int_part + dec_part
        else:
            return int(''.join(self._separated_digits))

    @staticmethod
    def strip_value(text: str):
        """Strip non-numeric characters from a string representing a number
        :param str text: Input string representing a number
        :return: String with stripped non-numeric characters
        :rtype str:
        """
        currencies = u'$¢£¤¥֏؋৲৳৻૱௹฿៛₠₡₢₣₤₥₦₧₨₩₪₫€₭₮₯₰₱₲₳₴₵₶₷₸₹₺₻₼₽₾꠸﷼﹩＄￠￡￥￦'
        text = text.strip(
            '\n\r\t ' + currencies + "'+*/%^_:;'\"\\|").rstrip('—-.,')
        text = text.replace(' ', '')
        return text

    def is_numeric(self) -> bool:
        """Determines if the text is a representation of a numeric value

        :return: True if the raw text corresponds to numeric representation
        """
        if re.match(self.numeric,
                    self.strip_value(self.raw)) and '%' not in self.raw:
            return True
        return False



"""This module defines the Document object,
which includes the input and output fields for the model,
as well as the OCR data and page images of the document"""
import copy
import itertools
import multiprocessing
import os
from collections import OrderedDict
from typing import Dict, List, Tuple, Any, Optional, Callable, Union

import fitz
import pandas as pd
from fitz.utils import getColor, getColorList

from pycognaize.document.html_info import HTML
from pycognaize.login import Login
from pycognaize.common.enums import IqDocumentKeysEnum, FieldTypeEnum
from pycognaize.common.field_collection import FieldCollection
from pycognaize.common.utils import cloud_interface_login
from pycognaize.document.field import FieldMapping, TableField
from pycognaize.document.field.field import Field
from pycognaize.document.page import Page
from pycognaize.document.tag import TableTag, ExtractionTag
from pycognaize.document.tag.cell import Cell
from pycognaize.document.tag.tag import BoxTag, LineTag


class Document:
    """Definition of input and output for a single document,
     depending on a given model"""

    def __init__(self,
                 input_fields: 'FieldCollection[str, List[Field]]',
                 output_fields: 'FieldCollection[str, List[Field]]',
                 pages: Dict[int, Page],
                 html_info: HTML,
                 metadata: Dict[str, Any]):
        self._login_instance = Login()
        self._metadata = metadata
        self._pages: Dict[int, Page] = pages if pages else None

        self._is_xbrl: bool = False
        self._html_info: HTML = html_info
        self._x: FieldCollection[str, List[Field]] = input_fields
        self._y: FieldCollection[str, List[Field]] = output_fields

    @property
    def x(self) -> 'FieldCollection[str, List[Field]]':
        """Returns a dictionary, where keys are input field names
        and values are list of Field objects"""
        return self._x

    @property
    def y(self) -> 'FieldCollection[str, List[Field]]':
        """Returns a dictionary, where keys are output field names
        and values are list of Field objects"""
        return self._y

    @property
    def metadata(self) -> Dict[str, Any]:
        """Returns document metadata"""
        return self._metadata

    @property
    def id(self) -> str:
        """Returns the pycognaize id of the document"""
        return self.metadata['document_id']

    @property
    def document_src(self):
        """Returns the source of the document"""
        return self.metadata['src']

    @property
    def pages(self) -> Dict[int, Page]:
        """Returns a dictionary, where each key is the page number
        and values are Page objects"""
        return self._pages

    @property
    def is_xbrl(self) -> bool:
        """Returns True if document is XBRL, otherwise False"""
        return self._html_info.path != ''

    @property
    def html(self):
        """Returns `HTML` object"""
        return self._html_info

    @staticmethod
    def get_matching_table_cells_for_tag(
            tag: BoxTag,
            table_tags: List[TableTag],
            one_to_one: bool
    ) -> List[Tuple[BoxTag, TableTag, Cell, float]]:
        """Create a list which includes the original extraction tag,
           the corresponding table tag and Cell objects
           and the IOU of the intersection

        :param tag: The `tag` for which matching
            table and cells should be found
        :param table_tags: List of `table_tag`s
        :param one_to_one: If true,
            for each tag only one corresponding cell will be returned
        :return: List of tuples,
            which include the original extraction tag,
            the corresponding table tag and Cell objects
            and the IOU of the intersection
        """
        intersection = []
        if isinstance(tag, TableTag):
            return []
        for ttag in table_tags:
            if ttag.page.page_number != tag.page.page_number:
                continue
            for cell in ttag.cells.values():
                temp_cell = copy.deepcopy(cell)
                temp_cell.page = ttag.page
                iou = tag.iou(temp_cell)
                if iou <= 0:
                    continue
                if one_to_one:
                    if (not intersection or (
                            intersection and intersection[0][-1] < iou
                    )):
                        intersection = [(tag, ttag, cell, iou)]
                else:
                    intersection.append((tag, ttag, cell, iou))
        return intersection

    def get_table_cell_overlap(
            self, source_field: str,
            one_to_one: bool) -> List[Tuple[BoxTag, TableTag, Cell, float]]:
        """Create a list which includes the original extraction tag,
           the corresponding table tag and Cell objects
           and the IOU of the intersection

        :param source_field: Name of the field,
            for which to return the corresponding table cells
        :param one_to_one: If true,
            for each tag only one corresponding cell will be returned
        :return: List of tuples,
            which include the original extraction tag,
            the corresponding table tag and Cell objects
            and the IOU of the intersection
        """
        # noinspection PyUnresolvedReferences
        table_tags = [
            tag
            for fields in itertools.chain(self.x.values(), self.y.values())
            for field in fields
            for tag in field.tags
            if isinstance(tag, TableTag)
        ]
        res = []
        if source_field in self.x:
            fields = self.x[source_field]
        elif source_field in self.y:
            fields = self.y[source_field]
        else:
            return []
        for field in fields:
            for tag in field.tags:
                intersection = self.get_matching_table_cells_for_tag(
                    tag=tag, table_tags=table_tags, one_to_one=one_to_one)
                res.extend(intersection)
        return res

    def get_tied_fields(self, tag: ExtractionTag,
                        field_type: str = FieldTypeEnum.BOTH.value,
                        threshold: float = 0.5,
                        pn_filter: Callable = lambda x: True
                        ) -> Dict[str, List[Field]]:
        """Given an `ExtractionTag`, return all the fields that contain
           tags in the same physical location.

        :param tag: Input `ExtractionTag`
        :param field_type: Types of fields to consider {input/output/both}
        :param threshold: The IoU threshold to consider the tags in the same
            location
        :param pn_filter: If provided, only fields with
            names passing the filter will be considered
        :return: Dictionary where key is pname
            and value is List of `Field` objects
        """
        all_tied_fields: Dict[str, List[Field]] = OrderedDict()
        if field_type == FieldTypeEnum.INPUT_FIELD.value:
            scopes = (self.x,)
        elif field_type == FieldTypeEnum.OUTPUT_FIELD.value:
            scopes = (self.y,)
        elif field_type == FieldTypeEnum.BOTH.value:
            scopes = (self.x, self.y)
        else:
            raise ValueError(
                f"'field_type' should be one of "
                f"{tuple(i.value for i in FieldTypeEnum.__members__.values())}"
                f" got {field_type}")
        for scope in scopes:
            for pname, fields in scope.items():
                if not pn_filter(pname):
                    continue
                tied_fields = {field for field in fields
                               if not isinstance(field, TableField)
                               for field_tag in field.tags
                               if isinstance(field_tag, ExtractionTag)
                               and (tag & field_tag)
                               / min({tag.area, field_tag.area}) >= threshold}
                if tied_fields:
                    all_tied_fields[pname] = list(tied_fields)
        return all_tied_fields

    def get_tied_tags(self, tag: ExtractionTag,
                      field_type: str = FieldTypeEnum.BOTH.value,
                      threshold: float = 0.9,
                      pn_filter: Callable = lambda x: True
                      ) -> Dict[str, List[ExtractionTag]]:
        """Given a single tag, return all other tags in the document that are
            in the same physical location in the document

        :param tag: Input `ExtractionTag`
        :param field_type: Types of fields to consider {input/output/both}
        :param threshold: The IoU threshold to consider the tags in the same
            location
        :param pn_filter: If provided, only tags that are in fields
            with names passing the filter will be considered
        :return: Dictionary where key is pname
            and value is List of `ExtractionTag` objects
        """
        all_tied_tags: Dict[str, List[ExtractionTag]] = OrderedDict()
        if field_type == FieldTypeEnum.INPUT_FIELD.value:
            scopes = (self.x,)
        elif field_type == FieldTypeEnum.OUTPUT_FIELD.value:
            scopes = (self.y,)
        elif field_type == FieldTypeEnum.BOTH.value:
            scopes = (self.x, self.y)
        else:
            raise ValueError(
                f"'field_type' should be one of "
                f"{tuple(i.value for i in FieldTypeEnum.__members__.values())}"
                f" got {field_type}")
        for scope in scopes:
            for pname, fields in scope.items():
                if not pn_filter(pname):
                    continue
                tied_tags = {field_tag for field in fields
                             if not isinstance(field, TableField)
                             for field_tag in field.tags
                             if isinstance(field_tag, ExtractionTag)
                             if tag.iou(field_tag) >= threshold}
                if tied_tags:
                    all_tied_tags[pname] = list(tied_tags)
        return all_tied_tags

    def get_first_tied_field(self, tag: ExtractionTag,
                             pn_filter: Callable = lambda x: True
                             ) -> Tuple[str, Field]:
        """Return the first field that is in the same location as the given tag

        :param tag: Input `ExtractionTag`
        :param pn_filter: If provided, only fields with
            names passing the filter will be considered
        :return: If match found, return Tuple of the matching pname
            and `Field`, otherwise return `None`
        """
        res = None
        fields = self.get_tied_fields(tag=tag,
                                      pn_filter=pn_filter)
        if fields:
            pname = list(fields)[0]
            first_tied_field = fields[pname][0]
            res = (pname, first_tied_field) if first_tied_field else res
        return res

    def get_first_tied_field_value(self, tag: ExtractionTag,
                                   pn_filter: Callable =
                                   lambda x: True):
        """Return the value of the first field that is in the
         same location as the given tag

        :param tag: Input `ExtractionTag`
        :param pn_filter: If provided, only tags that are
            in fields with names passing the filter will be considered
        :return:
        """
        if isinstance(tag, float):
            val = ''
        else:
            tied_field = self.get_first_tied_field(
                tag=tag, pn_filter=pn_filter)
            if tied_field is None:
                val = tag.raw_value
            else:
                pname, matching_field = self.get_first_tied_field(
                    tag=tag, pn_filter=pn_filter)
                # noinspection PyUnresolvedReferences
                val = matching_field.value
        return val

    def get_first_tied_tag(self, tag: ExtractionTag,
                           pn_filter: Callable = lambda x: True
                           ) -> Tuple[str, ExtractionTag]:
        """Return the first tag that is in the same location as the given tag

        :param tag: Input `ExtractionTag`
        :param pn_filter: If provided, only tags that
            are in fields with names passing the filter will be considered
        :return: If match found, return Tuple of the matching pname
            and `ExtractionTag`, otherwise return `None`
        """
        res = None
        tags = self.get_tied_tags(tag=tag,
                                  pn_filter=pn_filter)
        if tags:
            pname = list(tags)[0]
            first_tied_tag = tags[pname][0]
            res = (pname, first_tied_tag) if first_tied_tag else res
        return res

    def get_first_tied_tag_value(self, tag: ExtractionTag,
                                 pn_filter: Callable =
                                 lambda x: True):
        """Return the value of the first tag that is in the same
            location as the given tag

        :param tag: Input `ExtractionTag`
        :param pn_filter: If provided, only tags that are in
            fields with names passing the filter will be considered
        :return:
        """
        tied_tag = self.get_first_tied_tag(tag=tag,
                                           pn_filter=pn_filter)
        if tied_tag is None:
            val = tag.raw_value
        else:
            pname, matching_tag = tied_tag
            val = matching_tag.raw_value
        return val

    def get_df_with_tied_field_values(self, table_tag: TableTag,
                                      pn_filter: Callable =
                                      lambda x: True
                                      ) -> pd.DataFrame:
        """Return the dataframe of the TableTag, where each cell
           value is replaced with the values in the fields of
           tied values (e.i. values that are in
           the same physical location as the cell)

        :param table_tag: Input `TableTag`
        :param pn_filter: : If provided, only fields with
            names passing the filter will be considered
        :return: Dataframe of the TableTag
        """
        return table_tag.raw_df.applymap(
            lambda x: self.get_first_tied_field_value(
                x,
                pn_filter=pn_filter))

    def load_page_images(self, page_filter: Callable = lambda x: True) -> None:
        """Get all images of the pages in the document
         (Using multiprocessing)"""
        global _get_page

        def _get_page(page, filter_pages: Callable = page_filter):
            if filter_pages(page):
                _ = page.image_bytes
            return page

        pool = multiprocessing.Pool(min(multiprocessing.cpu_count() * 2, 16))
        pages = pool.map(_get_page, self.pages.values())
        self._pages = OrderedDict({page.page_number: page for page in pages})

    def load_page_ocr(self, page_filter: Callable = lambda x: True) -> None:
        """Get all OCR of the pages in the document
           (Using multiprocessing)"""
        global _get_page

        def _get_page(page, filter_pages: Callable = page_filter):
            if filter_pages(page):
                _ = page.ocr
            return page

        pool = multiprocessing.Pool(min(multiprocessing.cpu_count() * 2, 16))
        pages = pool.map(_get_page, self.pages.values())
        self._pages = OrderedDict({page.page_number: page for page in pages})

    def to_dict(self) -> dict:
        """Converts Document object to dict"""
        input_fields = OrderedDict(
            {name: [field.to_dict() for field in fields]
             for name, fields in self.x.items()})
        output_fields = OrderedDict(
            {name: [field.to_dict() for field in fields]
             for name, fields in self.y.items()})
        data = OrderedDict(input_fields=input_fields,
                           output_fields=output_fields,
                           metadata=self.metadata)
        return data

    @classmethod
    def from_dict(cls, raw: dict,
                  data_path: str) -> 'Document':
        """Document object created from data of dict
        :param raw: document dictionary
        :param data_path: path to the documents OCR and page images
        :param login_instance: login instance of pycognaize
        """
        if not isinstance(raw, dict):
            raise TypeError(
                f"Expected dict for 'raw' argument got {type(raw)} instead")
        metadata = raw['metadata']
        pages = OrderedDict()
        html_info = HTML(path=data_path, document_id=metadata['document_id'])
        for page_n in range(1, metadata['numberOfPages'] + 1):
            if (
                    'pages' in raw
                    and str(page_n) in raw['pages']
                    and 'width' in raw['pages'][str(page_n)]
                    and 'height' in raw['pages'][str(page_n)]
            ):
                image_width = raw['pages'][str(page_n)]['width']
                image_height = raw['pages'][str(page_n)]['height']
            else:
                image_width = None
                image_height = None
            pages[page_n] = Page(page_number=page_n,
                                 document_id=metadata['document_id'],
                                 path=data_path,
                                 image_width=image_width,
                                 image_height=image_height)
        input_fields = FieldCollection(
            {name: [
                FieldMapping[
                    field[IqDocumentKeysEnum.data_type.value]
                ].value.construct_from_raw(raw=field, pages=pages,
                                           html=html_info)
                for field in fields]
             for name, fields in raw['input_fields'].items()})
        output_fields = FieldCollection(
            {name: [
                FieldMapping[
                    field[IqDocumentKeysEnum.data_type.value]
                ].value.construct_from_raw(raw=field, pages=pages,
                                           html=html_info)
                for field in fields]
             for name, fields in raw['output_fields'].items()})
        return cls(input_fields=input_fields,
                   output_fields=output_fields,
                   pages=pages, html_info=html_info,
                   metadata=metadata)

    def _collect_all_tags_for_fields(self,
                                     field_names: List[str],
                                     is_input_field: bool = True)\
            -> List[Union[BoxTag, LineTag]]:
        """Collect all tags of given field names from either input or output
            fields

        :param field_names: List of strings representing the field names
        :param is_input_field: If true, collect tags from input fields,
            otherwise collect tags from output fields
        :return: List of tags from the specified fields
        """
        all_tags = []
        if is_input_field:
            field_dict = self.x
            field_type = 'input field'
        else:
            field_dict = self.y
            field_type = 'output field'
        if field_names is not None:
            for field_name in field_names:
                if field_name not in field_dict.keys():
                    raise ValueError(f'Invalid {field_type} {field_name}')
                for field in field_dict.get(field_name, []):
                    for tag in field.tags:
                        all_tags.append(tag)
        return all_tags

    def to_pdf(self,
               input_fields: Optional[List[str]] = None,
               output_fields: Optional[List[str]] = None,
               input_color: str = 'deeppink1',
               output_color: str = 'deepskyblue3',
               input_opacity: float = 0.2,
               output_opacity: float = 0.3) -> bytes:

        """
        Adds tags of input_fields and output_fields to the bytes object
        representing the pdf file of the document.

        :param input_fields: Input fields
        :param output_fields: Output fields
        :param input_color: The color of the annotation rectangle
            of the input field
        :param output_color: The color of the annotation rectangle
            of the output field
        :param input_opacity: The opacity of the annotation rectangle
            of the input field
        :param output_opacity: The opacity of the annotation rectangle
            of the output field
        :return: bytes object of the pdf
        """

        ci = cloud_interface_login(self._login_instance)
        pdf_path = os.path.join(self.pages[1].path, self.document_src) + '.pdf'

        with ci.open(pdf_path, 'rb') as f:
            pdf_bytes = f.read()
        doc_fitz = fitz.open('pdf', pdf_bytes)

        if input_fields is not None:
            input_tags = self._collect_all_tags_for_fields(
                field_names=input_fields, is_input_field=True)
            for tag in input_tags:
                pdf_bytes = annotate_pdf(
                    doc=doc_fitz,
                    tag=tag,
                    color=input_color,
                    opacity=input_opacity)
        if output_fields is not None:
            input_tags = self._collect_all_tags_for_fields(
                field_names=output_fields, is_input_field=False)
            for tag in input_tags:
                pdf_bytes = annotate_pdf(
                    doc=doc_fitz,
                    tag=tag,
                    color=output_color,
                    opacity=output_opacity)
        return pdf_bytes


def annotate_pdf(doc: fitz.Document,
                 tag: BoxTag,
                 color: str,
                 opacity: float = 0.3) -> bytes:
    """An annotated Document pdf in bytes"""
    page = doc[tag.page.page_number-1]
    x0 = tag.left * page.mediabox.width / 100
    y0 = tag.top * page.mediabox.height / 100
    x1 = tag.right * page.mediabox.width / 100
    y1 = tag.bottom * page.mediabox.height / 100
    annot_rect = fitz.Rect(x0, y0, x1, y1)
    if color.upper() not in getColorList():
        raise ValueError(f'Wrong color {color}')
    if opacity < 0 or opacity > 1:
        raise ValueError(f'Wrong opacity value {opacity}')
    color_dict = {"stroke": getColor(color), "fill": getColor(color)}
    annot = page.add_rect_annot(annot_rect)
    annot.set_colors(color_dict)
    annot.set_opacity(opacity)
    annot.update()
    return doc.write()



import logging
from collections import OrderedDict

from pycognaize.common.lazy_group_dict import LazyGroupDict
from pycognaize.document.field import Field


class FieldCollection(OrderedDict):
    """Contains fields included in the document by also providing
    functionality to group fields by group_name and group_key"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._groups = None
        self._groups_by_key = None

    @property
    def groups(self) -> dict:
        """Return Groups"""
        if self._groups is None:
            self._groups = LazyGroupDict(self).groups
        return self._groups

    @property
    def key_groups(self) -> dict:
        """Return Groups"""
        if self._groups_by_key is None:
            self._groups_by_key = LazyGroupDict(self).groups_by_key
        return self._groups_by_key

    def groups_by_key(self, group_key: str) -> dict:
        """Returns groups that contain the given group_key"""
        if group_key in self.key_groups.keys():
            return self.key_groups[group_key]
        else:
            raise KeyError

    def groups_by_name(self, name: str) -> dict:
        """Returns fields that are contained in group with the
        given group name"""
        if name in self.groups.keys():
            return self.groups[name]
        else:
            raise KeyError

    def groups_by_field(self, field: Field) -> list:
        """Returns groups that contain the given field"""
        name_group = self.groups_by_name(field.group_name)
        if field.group_key in name_group:
            return name_group[field.group_key]
        else:
            logging.warning(f"Field {field} is not in any group")
            return []



import bson
from typing import Union, Dict, TYPE_CHECKING

from pycognaize.common.enums import IqTagKeyEnum, ID
from pycognaize.document.tag.tag import LineTag

from pycognaize.common.utils import convert_coord_to_num
from pycognaize.document.tag.cell import Cell
if TYPE_CHECKING:
    from pycognaize.document.page import Page


class SectionTag(LineTag):
    """Represents field's coordinate data on document"""

    def __init__(self, top, page, tag_type):
        super().__init__(top=top, page=page, tag_type=tag_type)
        self._type = tag_type
        self._top = top
        self._page = page

    @classmethod
    def construct_from_raw(cls, raw: dict,
                           pages: Dict[int, 'Page'],
                           tag_type: str) -> 'SectionTag':
        """Builds Tag object from pycognaize raw data

        :param raw: pycognaize field's tag info
        :param pages: `Page` to which the tag belongs
        :param tag_type: `Page` to which the tag belongs
        :return:
        """
        page_number = raw['page']
        top = convert_coord_to_num(raw['top'])
        tag = cls(top=top, page=pages[page_number],
                  tag_type=tag_type)
        return tag

    def vshift(self, by) -> 'SectionTag':
        """Shifts line vertically

        :param by: the amount by which the tag should be vertically shifted
        :return: shifted rectangle
        """
        return self.__class__(top=self.top + by,
                              page=self.page,
                              tag_type=self.type)

    def vertical_shift(self, by):
        return self.vshift(by)

    def __add__(self, other: Union['LineTag', Cell]) -> 'SectionTag':
        """Merge two rectangles into one"""
        top = min(self.top, other.top)
        return SectionTag(top=top, page=self.page, tag_type=self.type)

    def to_dict(self) -> dict:
        """Converts extraction tag to dict"""
        return {
            ID: str(bson.ObjectId()),
            IqTagKeyEnum.top.value: f"{self.top}%",
            IqTagKeyEnum.page.value: self.page.page_number}



import math

import bson
from datetime import datetime
from typing import Union

from pycognaize.common.confidence import Confidence
from pycognaize.common.enums import IqTagKeyEnum, ID
from pycognaize.document.tag.tag import BoxTag

from pycognaize.common.utils import convert_coord_to_num
from pycognaize.document.tag.cell import Cell
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pycognaize.document.page import Page


class ExtractionTag(BoxTag):
    """Represents field's coordinate data on document"""

    def __init__(self, left, right, top, bottom, page, raw_value,
                 raw_ocr_value, confidence: Confidence = None):
        super().__init__(left=left, right=right, top=top, bottom=bottom,
                         page=page, confidence=confidence)
        self._raw_value = raw_value
        self._raw_ocr_value = raw_ocr_value

    @classmethod
    def construct_from_raw(cls, raw: dict, page: 'Page') -> 'ExtractionTag':
        """Builds Tag object from pycognaize raw data
        :param raw: pycognaize field's tag info
        :param page: `Page` to which the tag belongs
        :return:
        """
        confidence = Confidence(raw.get(IqTagKeyEnum.
                                confidence.value, {}))
        left = convert_coord_to_num(raw['left'])
        top = convert_coord_to_num(raw['top'])
        height = convert_coord_to_num(raw['height'])
        width = convert_coord_to_num(raw['width'])
        right = left + width
        bottom = top + height
        raw_value = raw['value']
        raw_ocr_value = raw['ocrValue']
        tag = cls(left=left, right=right, top=top, bottom=bottom,
                  page=page, raw_value=raw_value, raw_ocr_value=raw_ocr_value,
                  confidence=confidence)
        return tag

    def hshift(self, by) -> 'ExtractionTag':
        """Shifts rectangle horizontally

        :param by: the amount by which the tag should be horizontally shifted
        :return: shifted rectangle
        """
        return self.__class__(left=self.left + by, right=self.right + by,
                              top=self.top, bottom=self.bottom,
                              page=self.page, raw_value=self.raw_value,
                              raw_ocr_value=self.raw_ocr_value,
                              confidence=self.confidence)

    def horizontal_shift(self, by):
        return self.hshift(by)

    def vshift(self, by) -> 'ExtractionTag':
        """Shifts rectangle vertically
        :param by: the amount by which the tag should be vertically shifted
        :return: shifted rectangle
        """
        return self.__class__(left=self.left, right=self.right,
                              top=self.top + by, bottom=self.bottom + by,
                              page=self.page, raw_value=self.raw_value,
                              raw_ocr_value=self.raw_ocr_value,
                              confidence=self.confidence)

    def vertical_shift(self, by):
        return self.vshift(by)

    def __add__(self, other: Union['BoxTag', Cell]) -> 'ExtractionTag':
        """Merge two rectangles into one"""
        if self.page.page_number == other.page.page_number:
            left = min(self.left, other.left)
            right = max(self.right, other.right)
            top = min(self.top, other.top)
            bottom = max(self.bottom, other.bottom)
            raw_ocr_value_joined = " ".join(
                [i.raw_ocr_value for i in sorted(
                    [self, other], key=lambda x: (x.left, x.top))])
            return ExtractionTag(
                left=left, right=right, top=top, bottom=bottom,
                page=self.page, raw_value=self.raw_value,
                raw_ocr_value=raw_ocr_value_joined,
                confidence=self.confidence)
        else:
            raise ValueError("Tags are not on the same page.")

    @property
    def raw_value(self):
        return self._raw_value

    @property
    def raw_ocr_value(self):
        return self._raw_ocr_value

    def _validate_numeric(self):
        """Validate numerica data"""
        try:
            self.value = (float(self.raw_value) if self.raw_value is not None
                          else math.nan)
            self.has_value_exception = False
        except Exception as ValueException:
            self.has_value_exception = True
            self.value_exception_message = str(ValueException)

        try:
            self.ocr_value = (float(self.raw_ocr_value)
                              if self.raw_ocr_value is not None else math.nan)
            self.has_raw_value_exception = False
        except Exception as RawValueException:
            self.has_raw_value_exception = True
            self.raw_value_exception_message = str(RawValueException)

    def _validate_date(self, date_format):
        """Validate date data"""
        try:
            self.value = datetime.strptime(self.raw_value, date_format)
            self.has_value_exception = False
        except Exception as ValueException:
            self.has_value_exception = True
            self.value_exception_message = str(ValueException)

        try:
            self.ocr_value = datetime.strptime(self.raw_ocr_value, date_format)
            self.has_raw_value_exception = False
        except Exception as RawValueException:
            self.has_raw_value_exception = True
            self.raw_value_exception_message = str(RawValueException)

    def to_dict(self) -> dict:
        """Converts extraction tag to dict"""
        return {
            ID: str(bson.ObjectId()),
            IqTagKeyEnum.ocr_value.value: self.raw_ocr_value,
            IqTagKeyEnum.value.value: str(self.raw_value),
            IqTagKeyEnum.left.value: f"{self.left}%",
            IqTagKeyEnum.top.value: f"{self.top}%",
            IqTagKeyEnum.height.value: f"{self.bottom - self.top}%",
            IqTagKeyEnum.width.value: f"{self.right - self.left}%",
            IqTagKeyEnum.page.value: self.page.page_number,
            IqTagKeyEnum.confidence.value: self.confidence.get_confidence(),
        }



import enum

ID = "_id"

IMG_EXTENSION = 'jpeg'
OCR_DATA_EXTENSION = 'json'
SNAPSHOT_EXTENSION = 'pickle'


class FieldDataTypeEnum(enum.Enum):
    extraction = 'extraction'
    computation = 'computation'
    classification = 'classification'
    extraction_and_computation = 'extraction and computation'


class StorageEnum(enum.Enum):
    ocr_folder = 'data'
    image_folder = 'images'
    doc_file = 'document.json'
    pdf_file = 'original.pdf'
    html_file = 'source.html'
    # snap_file = f"snap.{SNAPSHOT_EXTENSION}"


class EnvConfigEnum(enum.Enum):
    IQ_DB_URL = "IQ_DB_URL"
    IQ_SNAP_STORAGE_PATH = "IQ_SNAP_STORAGE_PATH"
    IQ_DOCUMENTS_STORAGE_PATH = "IQ_DOCUMENTS_STORAGE_PATH"  # Images and OCR
    # IQ_DATASNAP_URL = "IQ_DATASNAP_URL"
    SNAPSHOT_ID = 'SNAPSHOT_ID'
    SNAPSHOT_PATH = "SNAPSHOT_PATH"  # Images, OCR and snap.json
    HOST = "API_HOST"


class IqCollectionEnum(enum.Enum):
    documents = 'documents'
    collections = 'collections'
    templates = 'templates'
    recipes = 'datarecipes'
    datasets = 'datasets'


class IqDatasetKeyEnum(enum.Enum):
    collections = 'collections'
    dataset_type = "type"
    recipe_id = "dataRecipeId"


class IqDatasetTypeEnum(enum.Enum):
    snapshot = "snapshot"
    pipeline = "pipeline"


class IqDocumentKeysEnum(enum.Enum):
    collection_id = 'collectionId'
    data_type = 'dataType'
    table = 'table'
    templates = 'templates'
    template = 'template'
    fields = 'fields'
    tags = 'tags'
    pages = 'pages'
    name = 'name'
    section = 'section'
    field_type = 'fieldType'
    field_id = '_id'
    src_field_id = "srcFieldId"
    page = 'page'
    src = 'src'
    page_src = 'src'


# TODO: move Document enum values to corresponding enum (field/tag etc)
class IqFieldKeyEnum(enum.Enum):
    data_type = 'dataType'
    field_type = 'fieldType'
    group = 'group'
    group_key = 'groupKey'
    has_valid_value = 'hasValidValue'
    multiple = 'multiple'
    name = 'name'
    order = 'order'
    repeatable = 'repeatable'
    required = 'required'
    src_field_id = 'srcFieldId'
    tags = 'tags'
    template_id = 'templateId'
    value = 'value'
    repeat_parent = 'repeatParent'


class IqTagKeyEnum(enum.Enum):
    value = 'value'
    has_valid_value = 'hasValidValue'
    top = 'top'
    left = 'left'
    width = 'width'
    height = 'height'
    page = 'page'
    ocr_value = 'ocrValue'
    has_valid_ocr_value = 'hasValidOcrValue'
    is_table = 'isTable'
    confidence = 'classConfidence'


class IqPageEnum(enum.Enum):
    page = 'page'
    src = 'src'


class IqDataTypesEnum(enum.Enum):
    table = 'table'
    text = 'text'
    date = 'date'
    number = 'number'
    area = 'area'


class IqRecipeEnum(enum.Enum):
    id = '_id'
    field_id = 'fieldId'
    name = 'name'
    templates = 'templates'
    template = 'template'
    template_id = 'template_id'
    fields = 'fields'
    python_name = 'pythonName'
    document_name = 'document_name'
    user = 'user'
    raw_field_type = 'rawFieldType'
    version = '__v'


class IqTableTagEnum(enum.Enum):
    table = 'table'
    cells = 'cells'
    page = 'page'
    left = "left"
    top = "top"
    height = "height"
    width = "width"
    confidence = 'classConfidence'


class IqCellKeyEnum(enum.Enum):
    left_col_top_row = 'left_col_top_row'
    col_span = "colspan"
    row_span = "rowspan"
    text = "value"
    left = "left"
    top = "top"
    width = "width"
    height = "height"


class PythonShellEnum(enum.Enum):
    zmq_interactive_shell = 'ZMQInteractiveShell'
    terminal_interactive_shell = 'TerminalInteractiveShell'
    other_type_shell = 'OtherTypeShell'
    standard_python_interpreter = 'StandardPythonInterpreter'


class FieldTypeEnum(enum.Enum):
    """Enumeration of field types"""
    INPUT_FIELD = 'input'
    OUTPUT_FIELD = 'output'
    BOTH = 'both'


class XBRLTagEnum(enum.Enum):
    """Enumeration for XBRL document fields"""
    source = 'source'
    td_id = '_id'
    xpath = 'xpath'
    anchor_id = 'anchorId'
    id = 'id'
    is_bold = 'isBold'
    left_indentation = 'leftIndentation'
    ids = 'ids'
    row_index = 'rowIndex'
    col_index = 'colIndex'
    tag_id = 'tagId'
    ocr_value = 'ocrValue'
    value = 'value'
    is_table = 'isTable'


class XBRLCellEnum(enum.Enum):
    id = 'id'
    source = 'source'
    ids = 'ids'
    html_id = 'id'
    xpath = 'xpath'
    row_index = 'rowIndex'
    col_index = 'colIndex'
    col_span = 'colspan'
    row_span = 'rowspan'
    raw_value = 'value'
    is_bold = 'isBold'
    left_indentation = 'leftIndentation'


class XBRLTableTagEnum(enum.Enum):
    source = 'source'
    ocr_value = 'ocrValue'
    value = 'value'
    is_table = 'isTable'
    is_bold = 'isBold'
    left_indentation = 'leftIndentation'
    anchor_id = 'anchorId'
    xpath = 'xpath'
    table = 'table'
    cells = 'cells'
    title = 'title'
    _id = '_id'



import logging
from typing import List, Optional, Dict, Type


from pycognaize.common.enums import (
    IqDocumentKeysEnum,
    ID,
    IqFieldKeyEnum,
    IqDataTypesEnum
)
from pycognaize.document.html_info import HTML
from pycognaize.document.page import Page
from pycognaize.document.field import Field
from pycognaize.document.tag.section_tag import SectionTag


class SectionField(Field):
    """Base class for all pycognaize Section fields"""
    tag_class: Type[SectionTag] = SectionTag

    def __init__(self,
                 name: str,
                 tags: Optional[List[SectionTag]] = None,
                 field_id: Optional[str] = None,
                 group_key: str = None,
                 confidence: Optional[float] = -1.0,
                 group_name: str = None
                 ):
        """ Create a SectionField object

        :param name: Name of the field
        :param value: Value of the field
            (left empty if values provided through tags)
        :param tags: List of tag objects
        :param field_id: The id of the field
        """
        super().__init__(name=name, tags=tags,
                         group_key=group_key, confidence=confidence,
                         group_name=group_name)
        self._field_id = field_id

    @classmethod
    def construct_from_raw(cls, raw: dict, pages: Dict[int, Page],
                           html: Optional[HTML] = None) -> 'SectionField':
        """Create SectionField object from dictionary"""
        section_dict: List[dict] = raw[IqDocumentKeysEnum.tags.value]
        tags = []
        for i in section_dict:
            for tag_type, tag_data in i[IqDocumentKeysEnum.
                                        section.value].items():
                try:
                    tags.append(cls.tag_class.construct_from_raw(
                                raw=tag_data,
                                pages=pages,
                                tag_type=tag_type))
                except Exception as e:
                    logging.debug(f"Failed creating tag"
                                  f" for field {raw[ID]}: {e}")
        return cls(name=raw[IqDocumentKeysEnum.name.value],
                   tags=tags,
                   field_id=str(raw[ID]),
                   group_key=raw.get(IqFieldKeyEnum.group_key.value, ''),
                   group_name=raw.get(IqFieldKeyEnum.group.value, '')
                   )

    def to_dict(self) -> dict:
        """Converts TextField object to dictionary"""
        field_dict = super().to_dict()
        field_dict[ID] = self._field_id
        field_dict[IqFieldKeyEnum.value.value] = self.value
        field_dict[IqFieldKeyEnum.data_type.value] = IqDataTypesEnum.text.value
        return field_dict

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name}>"

    def __str__(self):
        return f"<{self.__class__.__name__}: {self.name}>"



import io
import os
import re
import numpy as np
import bson
import logging

from PIL import Image
from itertools import groupby
from typing import Union, List, Optional, Iterable, Dict
from bson.json_util import loads as bson_loads
from dataclasses import dataclass
from cloudstorageio import CloudInterface

from pycognaize.login import Login
from pycognaize.common.enums import PythonShellEnum
from pycognaize.common.decorators import soon_be_deprecated


REGEX_NO_ALPHANUM_CHARS = re.compile(r'[^a-zA-Z\d)\[\](-.,]')


def is_float(str_number: str) -> bool:
    """Check if the string value is a valid number.
        Infinity and NaN are regarded as invalid"""
    # 'n' check covers both 'NaN', 'Inf' and 'Infinity'
    if 'n' in str_number.lower():
        return False
    try:
        float(str_number)
        return True
    except ValueError:
        return False


def convert_coord_to_num(val: Union[float, int, str]) -> Union[int, float]:
    """If the input is a string representation of a number
        (with or without a %), convert it into float.
        If the input is already a float or an integer, return itself"""
    if isinstance(val, str):
        val = val.rstrip('%')
        if not is_float(val):
            raise ValueError(f"Invalid numeric {val}")
        val = float(val)
    elif isinstance(val, (float, int)):
        pass
    else:
        raise TypeError(f"Invalid type for numeric {type(val)} ({val}")
    return val


def load_bson_by_path(doc_path):
    with open(os.path.join(doc_path), 'r') as f:
        return bson_loads(f.read())


@soon_be_deprecated()
def bytes_to_array(img_str: bytes) -> np.ndarray:
    """Convert image bytes into numpy array

    :param img_str: Bytestream of the image
    :return: numpy array of the image
    """
    import cv2
    nparr = np.frombuffer(img_str, np.uint8)
    img_np = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    return img_np


@soon_be_deprecated()
def string_to_array(img_str: bytes):
    """
    Convert a bytestring into a numpy array with opencv

    :param str img_str: image as byte string
    :return: img_np: image as numpy array
    """
    import cv2
    nparr = np.frombuffer(img_str, np.uint8)
    img_np = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    assert type(img_np) == np.ndarray, (
        "Expected numpy array, received {}".format(type(img_np)))

    return img_np


def image_bytes_to_array(img_str: bytes) -> np.ndarray:
    """Convert image bytes into numpy array

    :param img_str: Bytestream of the image
    :return: numpy array of the image
    """
    image = Image.open(io.BytesIO(img_str))
    # Remove <alpha> channel if it exists
    img_np = np.array(image)[..., :3]

    return img_np


def image_string_to_array(img_str: bytes):
    """
    Convert a bytestring into a numpy array with opencv

    :param str img_str: image as byte string
    :return: img_np: image as numpy array
    """
    image = Image.open(io.BytesIO(img_str))
    # Remove <alpha> channel if it exists
    img_np = np.array(image)[..., :3]
    assert type(img_np) == np.ndarray, (
        "Expected numpy array, received {}".format(type(img_np)))

    return img_np


def compute_otsu_threshold(img_array):
    bins_num = 256
    hist, bin_edges = np.histogram(img_array, bins=bins_num)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2.

    weight1 = np.cumsum(hist)
    weight2 = np.cumsum(hist[::-1])[::-1]

    mean1 = np.cumsum(hist * bin_centers) / weight1
    mean2 = (np.cumsum((hist * bin_centers)[::-1]) / weight2[::-1])[::-1]

    weight_mul = weight1[:-1] * weight2[1:]
    inter_class_variance = weight_mul * (mean1[:-1] - mean2[1:]) ** 2
    index_of_max_val = np.argmax(inter_class_variance)

    threshold = bin_centers[:-1][index_of_max_val]

    return threshold


def img_to_black_and_white(img_array):
    """Image to only black and white image

    black pixel --> white pixel
    other pixel --> black pixel

    :return:
    """

    def grayscale2bw(x, threshold_value):
        return 255 if x < threshold_value else 0

    vectorized_converter = np.vectorize(grayscale2bw)
    grayscale = np.dot(img_array[..., :3], [0.299, 0.587, 0.114])
    threshold = compute_otsu_threshold(img_array)
    return vectorized_converter(grayscale, threshold)


def group_sequence(list_of_integers):
    """Group sequence into list
    example ->
    test_list = [1, 2, 3, 8, 15, 23, 24, 25, 10, 11, 13, 15]
    sequence_list = [[1, 2, 3], [23, 24, 25], [10, 11]]
    :param list_of_integers: list of integers
    :return: list of grouped integers
    """
    it = iter(list_of_integers)
    prev, res = next(it), []

    while prev is not None:
        start = next(it, None)
        if prev + 1 == start:
            res.append(prev)
        elif res:
            yield list(res + [prev])
            res = []
        prev = start


def stick_word_boxes(box_coord: List[dict], img_bytes: bytes, padding=1):
    """Stick boxes with OpenCV
        1) Stick by y (height), add "2 pixel" padding in bottom and right
            Get bigger sequence of 255 < vectors_y < 0  --> bigger_sequence
            new_top = min(bigger_sequence)
            new_bottom = max(bigger_sequence)

        2) stick by x (width), add "2 pixel" padding in bottom
            Get numpy array of 255 < vectors_x < 0  --> min_max_x
            new_left = min(min_max_x)
            new_right = max(min_max_x)

    :param box_coord: List of dictionaries representing
        boxes with the following format:
        ({'left': <int>, 'right': <int>, 'top': <int>, 'bottom': <int>})
    :param img_bytes: The original image to which the coordinates correspond to
    :param padding: Expand the final boxes with this amount of pixels
        on each four sides
    :return:
    """
    np_image = image_bytes_to_array(img_bytes)
    b_and_w_image = img_to_black_and_white(np_image)
    h, w, _ = np_image.shape
    for word in box_coord:
        start_point_xmin = max(0, int(round(word['left'])) - 1)
        start_point_ymin = max(0, int(round(word['top'])) - 1)

        end_point_xmax = int(round(word['right'])) + 2
        end_point_ymax = int(round(word['bottom'])) + 2

        cropped_word = b_and_w_image[start_point_ymin:end_point_ymax,
                                     start_point_xmin:end_point_xmax]

        vectors_y = np.mean(cropped_word, axis=1)
        min_max_y = np.where((vectors_y != 0) & (vectors_y != 255))

        if len(min_max_y[0]) == 0:
            continue
        elif 0 < len(min_max_y[0]) < 3:
            vectors_x = np.mean(cropped_word, axis=0)
            min_max_x = np.where((vectors_x != 0) & (vectors_x != 255))

            new_top = min(min_max_y[0]) - 2 if len(min_max_y[0]) else 0
            new_bottom = max(min_max_y[0]) + 2 if len(min_max_y[0]) else 0
            new_left = min(min_max_x[0]) - 2 if len(min_max_x[0]) else 0
            new_right = max(min_max_x[0]) + 2 if len(min_max_x[0]) else 0
        else:
            try:
                sequence_y = list(group_sequence(list(min_max_y[0])))
                if len(sequence_y) > 0:
                    sequence_y = sorted(sequence_y, key=lambda x: len(x),
                                        reverse=True)
                max_sequence = sequence_y[0]
                new_top = min(max_sequence)
                new_bottom = max(max_sequence)
                new_bottom_cropped = new_bottom + 2

                cropped_word_height = cropped_word[
                                      new_top:new_bottom_cropped, :]

                vectors_x_height_stuck = np.mean(cropped_word_height, axis=0)
                min_max_x_height_stuck = np.where(
                    (vectors_x_height_stuck != 0
                     ) & (vectors_x_height_stuck != 255)
                )

                new_left = int(min(min_max_x_height_stuck[0]))
                new_right = int(max(min_max_x_height_stuck[0]))
            except Exception as e:
                logging.debug(f"Failed for word {word}: {e}")
                continue

        word['left'] = int(new_left) + start_point_xmin - padding
        word['right'] = int(new_right) + start_point_xmin + padding
        word['top'] = int(new_top) + start_point_ymin - padding
        word['bottom'] = int(new_bottom) + start_point_ymin + padding

    return box_coord


def preview_img(img: np.ndarray, size: int = 1000):
    """Preview the given image in a window

    :param img: Original image as numpy array
    :param size: Resize the window to the given resolution before viewing
    """
    import matplotlib.pyplot as plt
    plt.imshow(img)
    fig = plt.figure(num=1)
    fig.set_dpi(size)
    fig.set_size_inches(1, 1, forward=True)
    plt.axis('off')
    plt.show()


def infer_rows_from_words(box, class_ocr_data, auto_thresh=True, thresh=12,
                          min_height=10, min_width=10):
    """
    Infer row coordinates by class crop (table, column, row)
        from the ocr output
    :param class_ocr_data:
    :param box:
    :param auto_thresh:
    :param thresh:
    :param min_height:
    :param min_width:
    :rtype: tuple(list, list)
    :return:
    """
    class_ocr_data = sorted(class_ocr_data,
                            key=lambda x: (x['top'], x['left']))
    class_ocr_data = [
        i for i in class_ocr_data
        if i['ocr_text'].strip(' \t\n.,:;*_|\\/!\'"')]
    if auto_thresh:
        cleaned_ocr_data_for_height_detection = [
            i for i in class_ocr_data if
            i['ocr_text'].strip(' \t\n.,:;*_-~=|\\/!\'"')]
        all_heights_tuples = [
            (i.get('bottom') - i.get('top'), i['ocr_text']) for i in
            cleaned_ocr_data_for_height_detection
            if i.get('bottom') and i.get('top')]
        all_heights = [i[0] for i in all_heights_tuples]
        if len(all_heights) > 0:
            average_heights = sum(all_heights) / len(all_heights)
        else:
            average_heights = 0
        thresh = int(round(average_heights * 0.8))
        min_height = int(round(average_heights * 0.7))
        min_width = int(round(average_heights * 0.7))
    previous_bottom = 0
    previous_top = 0
    word_groups = []
    # Group aligned rows
    for i, (row_top, row_boxes) in enumerate(
            groupby(class_ocr_data, key=lambda x: x.get('top'))):
        word_boxes = sorted(list(row_boxes), key=lambda x: (x.get('left')))
        row_bottom = max(i['bottom'] for i in word_boxes)
        if not i:
            word_groups.append(word_boxes)
        elif (
                # (row_top - previous_bottom < thresh) or
                (abs(row_top - previous_top)
                 < thresh and abs(row_bottom - previous_bottom) < thresh) or
                (row_top >= previous_top and row_bottom <= previous_bottom) or
                (row_top <= previous_top and row_bottom >= previous_bottom)
        ):
            word_groups[-1] += word_boxes
        else:
            word_groups.append(word_boxes)
        previous_bottom = row_bottom
        previous_top = row_top
    aggregated_rows = []
    # Construct row boxes from row groups
    for row in word_groups:
        temp_d = dict(left=min(i['left'] for i in row) - box['left'],
                      right=max(i['right'] for i in row) - box['left'],
                      top=min(i['top'] for i in row) - box['top'],
                      bottom=max(i['bottom'] for i in row) - box['top'],
                      ocr_text=' '.join(i['ocr_text'] for i in row)
                      )
        if (temp_d['bottom'] - temp_d['top'] > min_width
                and temp_d['right'] - temp_d['left'] > min_height):
            aggregated_rows.append(temp_d)
    return aggregated_rows, word_groups


def clean_ocr_data(ocr_data: dict, thresh: float = 4.0) -> dict:
    """Cleans the ocr data

    :param ocr_data: OCR data dictionary, where values under the key `words`
        have the following form: {'top'}
    :param thresh: Words that
    :return:
    """
    ocr_data['words'] = [i for i in ocr_data['words'] if i['ocr_text'].strip()
                         and (i['bottom'] - i['top'])
                         / max(i['right'] - i['left'], 0.01) <= thresh]
    return ocr_data


def find_first_word_coords(text: str, ocr_data: list,
                           case_sensitive: bool = False, sort: bool = False,
                           clean: bool = True,
                           cleanup_regex=REGEX_NO_ALPHANUM_CHARS
                           ) -> Optional[dict]:
    """
    Detect the coordinates of the first occurrence
        of `text` in `ocr_data` if any.
    If the `text` is not found in `ocr_data` return None
    :param text:
    :param ocr_data: List of dictionaries.
    Each dictionary contains information about a single word.
    Each word dictionary has the following keys: `confidence`, `right`,
        `left`, `top`, `bottom`, `ocr_text`, `word_id_number`
    :param case_sensitive: If True, the search will be case-sensitive
    :param sort: If True, ocr_data will be ordered by `word_id_number`
        key before searching
    :param clean: If true, disregard all non-alphanumeric character
        from the search
    :param re._pattern_type cleanup_regex: Optional.
        Provide the regex for cleanup to be used
        (has effect only if `clean=True`)
    :return: Dictionary with word coordinates
        (keys: `left`, `right`, `top`, `bottom`, `matched_words`.
        `matched_words` includes the original word coordinate data
        for the matched words)
    """
    matched_words = []
    if sort:
        ocr_data = sorted(ocr_data, key=lambda x: float(x['word_id_number']))
    if not case_sensitive:
        text = text.lower()
    words = text.split(' ')
    idx = 0
    for ocr_d in ocr_data:
        ocr_word = ocr_d['ocr_text']
        if clean:
            ocr_word = cleanup_regex.sub('', ocr_word)
            word = cleanup_regex.sub('', words[idx])
        else:
            word = words[idx]
        ocr_word = ocr_word.strip()
        if not case_sensitive:
            ocr_word = ocr_word.lower()
        if ocr_word == word:
            idx += 1
            matched_words.append(ocr_d)
            if idx >= len(words):
                break
        else:
            matched_words = []
            idx = 0
    else:
        return
    return dict(top=min(matched_words, key=lambda x: x['top'])['top'],
                bottom=max(matched_words, key=lambda x: x['bottom'])['bottom'],
                left=min(matched_words, key=lambda x: x['left'])['left'],
                right=max(matched_words, key=lambda x: x['right'])['right'],
                matched_words=matched_words)


def intersects(word: dict,
               left: [int, float],
               right: [int, float],
               top: [int, float],
               bottom: [int, float]) -> bool:
    """
    Given the word and coordinates of the area,
        returns true if word intersects area, otherwise false
    :param word: dictionary, containing coordinates,
        ocr_text and word_id_number of the word
    :param left: left coordinate
    :param right: right coordinate
    :param top: top coordinate
    :param bottom: bottom coordinate
    :return: bool
    """
    return (word['left'] < right and word['right'] > left
            and word['top'] < bottom and word['bottom'] > top)


def compute_intersection_area(word: dict,
                              left: [int, float],
                              right: [int, float],
                              top: [int, float],
                              bottom: [int, float]) -> [int, float]:
    """
    Given the word and coordinates of the area, computes area of intersection
    :param word: dictionary, containing coordinates,
        ocr_text and word_id_number of the word
    :param left: left coordinate
    :param right: right coordinate
    :param top: top coordinate
    :param bottom: bottom coordinate
    :return: area of intersection
    """
    left = max([left, word['left']])
    top = max([top, word['top']])
    right = min([right, word['right']])
    bottom = min([bottom, word['bottom']])
    return (right - left) * (bottom - top)


def detect_python_shell() -> str:
    """Detect what python shell is being used - Standard Interpreter,
        Jupyter Notebook, Interactive shell or other"""
    try:
        # noinspection PyUnresolvedReferences
        shell = get_ipython().__class__.__name__
        if shell == PythonShellEnum.zmq_interactive_shell.value:
            # Jupyter notebook or qtconsole
            return PythonShellEnum.zmq_interactive_shell.value
        elif shell == PythonShellEnum.terminal_interactive_shell.value:
            # Terminal running IPython
            return PythonShellEnum.terminal_interactive_shell.value
        else:
            # Other type (?)
            return PythonShellEnum.other_type_shell.value
    except NameError:
        # Probably standard Python interpreter
        return PythonShellEnum.standard_python_interpreter.value


def replace_object_ids_with_string(bson_obj):
    if isinstance(bson_obj, dict):
        for key, value in bson_obj.items():
            bson_obj[key] = replace_object_ids_with_string(value)
    if isinstance(bson_obj, list):
        for idx, el in enumerate(bson_obj):
            bson_obj[idx] = replace_object_ids_with_string(el)
    if isinstance(bson_obj, bson.ObjectId):
        return str(bson_obj)
    else:
        return bson_obj


def empty_keys(obj: Union[List, Dict, Iterable], keys: List[str],
               empty: bool = False):
    if empty:
        return ''
    if isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = empty_keys(value, keys=keys, empty=key in keys)
    if isinstance(obj, list):
        for idx, el in enumerate(obj):
            obj[idx] = empty_keys(el, keys=keys)
    return obj


@dataclass
class ConfusionMatrix:
    TP: int = 0
    TN: int = 0
    FP: int = 0
    FN: int = 0

    @property
    def f1(self) -> float:
        try:
            return self.TP / (self.TP + 0.5 * (self.FP + self.FN))
        except ZeroDivisionError:
            return 0

    @property
    def precision(self) -> float:
        try:
            return self.TP / (self.TP + self.FP)
        except ZeroDivisionError:
            return 0

    @property
    def recall(self) -> float:
        try:
            return self.TP / (self.TP + self.FN)
        except ZeroDivisionError:
            return 0

    @property
    def accuracy(self) -> float:
        try:
            return (self.TP + self.TN) / (
                    self.FP + self.FN + self.TP + self.TN)
        except ZeroDivisionError:
            return 0

    def to_dict(self) -> dict:
        return {
            'TP': self.TP,
            'TN': self.TN,
            'FP': self.FP,
            'FN': self.FN
        }

    def compute_metrics(self) -> dict:
        """ Given confusion matrix, computes f1 score, precision,
            recall and accuracy.
         :return: results in dict"""
        result = {
            'confusion matrix': self.to_dict(),
            'f1 score': self.f1,
            'precision': self.precision,
            'recall': self.recall,
            'accuracy': self.accuracy}
        return result

    def draw_cm(self, title: str = 'Confusion Matrix'):
        cf_matrix = [self.TP, self.FP, self.FN, self.TN]
        cf_matrix = np.asarray(cf_matrix).reshape(2, 2)
        group_names = ['True Positive', 'False Positive',
                       'False Negative', 'True Negative']
        group_counts = ["{0:0.0f}".format(value)
                        for value in cf_matrix.flatten()]
        group_percentages = [
            "{0:.2%}".format(value)
            for value in cf_matrix.flatten() / np.sum(cf_matrix)]
        labels = [
            f'{v1}\n{v2}\n{v3}'
            for v1, v2, v3 in zip(group_names,
                                  group_counts,
                                  group_percentages)]
        labels = np.asarray(labels).reshape(2, 2)
        import seaborn as sns
        import matplotlib.pyplot as plt

        plt.figure(figsize=(12, 7))
        plt.title(title, fontsize=20)
        sns.heatmap(
            cf_matrix,
            annot=labels,
            fmt='',
            xticklabels=[1, 0],
            yticklabels=[1, 0],
            cmap='Blues',
            annot_kws={"size": 17})
        plt.show()


def get_index_of_first_non_empty_list(list_of_lists):
    non_empty_idx = len(list_of_lists)
    for idx, line in reversed(list(enumerate(list_of_lists))):
        if line:
            return idx
    return non_empty_idx


def filter_out_nested_lines(lines: List[str]) -> List[str]:
    """ Filters out nested html text lines"""
    lines_copy = lines.copy()
    sorted_lines = sorted(lines, key=lambda x: len(x))
    for idx, line in enumerate(sorted_lines):
        for next_line in sorted_lines[idx + 1:]:
            if line in next_line:
                lines_copy.remove(line)
                break
    return lines_copy


def convert_tag_coords_to_percentages(tag, w, h) -> dict:
    return dict(left=tag.left * w / 100,
                right=tag.right * w / 100,
                top=tag.top * h / 100,
                bottom=tag.bottom * h / 100)


def cloud_interface_login(login_instance: Login) -> CloudInterface:
    """Logs in to cloud interface"""

    if login_instance.logged_in:
        ci_instance = CloudInterface(
            aws_access_key_id=login_instance.aws_access_key,
            aws_secret_access_key=login_instance.aws_secret_access_key,
            aws_session_token=login_instance.aws_session_token)
    else:
        ci_instance = CloudInterface()
    return ci_instance



"""This module provides login functionality for downloading snapshots"""
import os
import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from pycognaize.common.enums import EnvConfigEnum
from pycognaize.common.exceptions import ServerAPIException


class Login:
    """Allows to access the AWS S3 bucket using cognaize credentials"""
    INSTANCE = None
    _access_key = ''
    _secret_access_key = ''
    _session_token = ''
    _snapshot_root = ''
    _logged_in = False

    def __new__(cls) -> 'Login':
        if cls.INSTANCE is None:
            cls.INSTANCE = super().__new__(cls)
        return cls.INSTANCE

    @property
    def logged_in(self) -> bool:
        """Return `True` if already logged in"""
        return self._logged_in

    @property
    def aws_access_key(self) -> str:
        """Access key for AWS"""
        return self._access_key

    @property
    def aws_secret_access_key(self) -> str:
        """Secret access key for AWS"""
        return self._secret_access_key

    @property
    def aws_session_token(self) -> str:
        """Session token for AWS"""
        return self._session_token

    @property
    def snapshot_root(self) -> str:
        """Root path for snapshots"""
        return self._snapshot_root

    def login(self, email: str, password: str):
        """Get AWS access credentials and stores in the instance"""
        retry_strategy = Retry(
            total=3,
            status_forcelist=[500, 502, 503, 504, 404],
            allowed_methods=["POST"]
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        http = requests.Session()
        http.mount("http://", adapter)
        http.mount("https://", adapter)

        host = os.environ.get(EnvConfigEnum.HOST.value)
        if host is None:
            raise EnvironmentError(
                f"{EnvConfigEnum.HOST.value} environment variable"
                f" required for login"
            )
        url = f"{host}/api/v1/integration/storage/token"

        authentication = {'email': email,
                          'password': password}

        try:
            user_credentials_response = http.post(url,
                                                  json=authentication,
                                                  timeout=10)
        except requests.exceptions.ConnectionError:
            raise ServerAPIException(f'Failed connecting to url: {url}')
        except requests.exceptions.Timeout:
            raise ServerAPIException(f'Connection timed out: {url}')

        print(user_credentials_response)
        user_credentials = user_credentials_response.json()
        if user_credentials_response.status_code == 200:
            self._logged_in = True
            self._access_key = user_credentials['credentials']['AccessKeyId']
            self._secret_access_key = \
                user_credentials['credentials']['SecretAccessKey']
            self._session_token = \
                user_credentials['credentials']['SessionToken']
            self._snapshot_root = user_credentials['snapshotRoot']
        elif user_credentials_response.status_code == 403:
            raise ServerAPIException("Data download permission error. "
                                     "Please make sure you have access"
                                     " to Snapshots")
        elif user_credentials_response.status_code == 401:
            raise ServerAPIException("Wrong email or password. "
                                     "Please make sure you entered the "
                                     "correct credentials")
        else:
            raise ServerAPIException("Server error. There was a problem "
                                     "with the serve")

    @classmethod
    def destroy(cls) -> None:
        """Delete singleton"""
        cls.INSTANCE = None



"""
| TextField objects represent a single labeled pycognaize field.
| The latter can have no tags, a single tag or multiple tags.

>>> from pycognaize.document.field.text_field import TextField
>>> text_field = TextField(name='company name', value='gooloogooloo')
>>> text_field.name
'company name'
>>> text_field.value
'gooloogooloo'
>>> text_field.tags
[]
"""
import logging
from typing import List, Optional, Dict, Type


from pycognaize.common.enums import (
    IqDocumentKeysEnum,
    IqTagKeyEnum,
    ID,
    IqFieldKeyEnum,
    IqDataTypesEnum
)
from pycognaize.document.html_info import HTML
from pycognaize.document.page import Page
from pycognaize.document.field import Field
from pycognaize.document.tag import ExtractionTag
from pycognaize.document.tag.html_tag import TDTag


class TextField(Field):
    """Base class for all pycognaize text fields"""
    tag_class: Type[ExtractionTag] = ExtractionTag
    html_tag_class: Type[TDTag] = TDTag

    def __init__(self,
                 name: str,
                 value: str = '',
                 tags: Optional[List[ExtractionTag]] = None,
                 field_id: Optional[str] = None,
                 group_key: str = None,
                 confidence: Optional[float] = -1.0,
                 group_name: str = None
                 ):
        """ Create a TextField object

        :param name: Name of the field
        :param value: Value of the field
            (left empty if values provided through tags)
        :param tags: List of tag objects
        :param field_id: The id of the field
        """
        super().__init__(name=name, tags=tags, value=value,
                         group_key=group_key, confidence=confidence,
                         group_name=group_name)
        self._field_id = field_id
        self._value = '; '.join([i.raw_value
                                 for i in self.tags]) if self.tags else value

    @property
    def value(self):
        return self._value

    @classmethod
    def construct_from_raw(cls, raw: dict, pages: Dict[int, Page],
                           html: Optional[HTML] = None) -> 'TextField':
        """Create TextField object from dictionary"""
        tag_dicts: List[dict] = raw[IqDocumentKeysEnum.tags.value]
        tags = []
        for i in tag_dicts:
            try:
                if pages:
                    tags.append(cls.tag_class.construct_from_raw(
                        raw=i, page=pages[i['page']]))
                else:
                    tags.append(cls.html_tag_class.construct_from_raw(
                        raw=i, html=html))
            except Exception as e:
                logging.debug(f"Failed creating tag for field {raw[ID]}: {e}")
        value = (tags[0].raw_value if (html.path and tags)
                 else raw[IqTagKeyEnum.value.value])
        return cls(name=raw[IqDocumentKeysEnum.name.value],
                   value=value,
                   tags=tags,
                   field_id=str(raw[ID]),
                   group_key=raw.get(IqFieldKeyEnum.group_key.value, ''),
                   group_name=raw.get(IqFieldKeyEnum.group.value, '')
                   )

    def to_dict(self) -> dict:
        """Converts TextField object to dictionary"""
        field_dict = super().to_dict()
        field_dict[ID] = self._field_id
        field_dict[IqFieldKeyEnum.value.value] = self.value
        field_dict[IqFieldKeyEnum.data_type.value] = IqDataTypesEnum.text.value
        return field_dict

    def __repr__(self):
        return (f"<{self.__class__.__name__}: {self.name}:"
                f" {'|'.join([i.raw_value for i in self.tags])}>")

    def __str__(self):
        return f"{'|'.join([i.raw_value for i in self.tags])}"



from typing import Optional, Dict, List, Type
import logging

from pycognaize.common.enums import (
    IqDocumentKeysEnum,
    ID,
    IqDataTypesEnum,
    IqFieldKeyEnum
)

from pycognaize.document.field import Field
from pycognaize.document.html_info import HTML
from pycognaize.document.tag.span_tag import SpanTag
from pycognaize.document.page import Page


class SpanField(Field):
    """Base class for all pycognaize table fields"""
    tag_class: Type[SpanTag] = SpanTag

    def __init__(self,
                 name: str,
                 tags: Optional[SpanTag] = None,
                 field_id: Optional[str] = None,
                 group_key: str = None,
                 confidence: Optional[float] = -1.0,
                 group_name: str = None,
                 ):
        tags = [] if tags is None else tags
        super().__init__(name=name, tags=tags,
                         group_key=group_key, confidence=confidence,
                         group_name=group_name)
        self._field_id = field_id

    @classmethod
    def construct_from_raw(cls, raw: dict, pages: Dict[int, Page],
                           html: Optional[HTML] = None) -> 'SpanField':
        """Create SnapField object from dictionary"""
        tag_dicts: List[dict] = raw[IqDocumentKeysEnum.tags.value]
        tags = []
        for i in tag_dicts:
            try:
                tags.append(cls.tag_class.construct_from_raw(
                    raw=i, page=pages[i['page']]))
            except Exception as e:
                logging.debug(f"Failed creating tag for field {raw[ID]}: {e}")
        return cls(name=raw[IqDocumentKeysEnum.name.value],
                   tags=tags[0] if tags else None,
                   field_id=str(raw[ID]),
                   group_key=raw.get(IqFieldKeyEnum.group_key.value, ''),
                   group_name=raw.get(IqFieldKeyEnum.group.value, '')
                   )

    def to_dict(self) -> dict:
        """Converts TableField object to dictionary"""
        field_dict = super().to_dict()
        field_dict[ID] = self._field_id
        field_dict[
            IqFieldKeyEnum.data_type.value] = IqDataTypesEnum.table.value
        field_dict[IqFieldKeyEnum.value.value] = ''
        return field_dict

    def order_tags(self):
        """Order tags by pdf data"""
        raise NotImplementedError

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name}>"

    def __str__(self):
        return self.__repr__()



import abc
from typing import Optional, List

import bson
import pandas as pd

from pycognaize.common.enums import (IqRecipeEnum, XBRLCellEnum,
                                     XBRLTableTagEnum, XBRLTagEnum, ID)
from pycognaize.document.html_info import HTML
from pycognaize.document.tag.tag import Tag


class HTMLTag(Tag, metaclass=abc.ABCMeta):
    """Base class for XBRL document tags"""
    def __init__(self, html_id: List[str], xpath: str,
                 tag_id: Optional[str] = None):
        self._html_id = html_id
        self._xpath = xpath
        self._tag_id = tag_id

    @property
    def html_id(self):
        return self._html_id

    @property
    def xpath(self):
        return self._xpath

    @property
    def tag_id(self):
        return self._tag_id

    @classmethod
    def construct_from_raw(cls, raw: dict, html: HTML) -> 'HTMLTag':
        ...


class HTMLTableTag(HTMLTag):
    """Represents table's coordinate data in XBRL document"""

    def __init__(self, tag_id: str,  value: str, ocr_value: str,
                 xpath: str, title: str, html_id: List[str], cell_data: dict,
                 html: HTML, source_ids, is_table: bool = True):
        super().__init__(html_id=html_id, xpath=xpath, tag_id=tag_id)
        self._value = value
        self._ocr_value = ocr_value
        self._is_table = is_table
        self._title = title
        self._cell_data = cell_data
        self._cells = {}
        self._html = html
        self._source_ids = source_ids
        self._populate_cells()
        self._build_df()
        self._raw_df = None
        self._df = None

    @property
    def title(self):
        return self._title

    @property
    def source_ids(self):
        return self._source_ids

    @property
    def value(self):
        return self._value

    @property
    def ocr_value(self):
        return self._ocr_value

    @property
    def is_table(self):
        return self._is_table

    @property
    def html(self):
        return self._html

    @property
    def cells(self):
        return self._cells

    @property
    def cell_data(self) -> dict:
        if not self._cell_data:
            raise Exception('Cell data is empty')
        return self._cell_data

    @property
    def raw_df(self) -> pd.DataFrame:
        if self._raw_df is None:
            self._raw_df = self._build_df()
        return self._raw_df

    @property
    def df(self) -> pd.DataFrame:
        if self._df is None:
            self._df = self.raw_df.applymap(
                lambda x: self._extract_value(x))
        return self._df

    @staticmethod
    def _extract_value(x):
        """Returns text value from `TDTag` object"""
        try:
            return x.raw_value
        except AttributeError:
            return ''

    def _populate_cells(self):
        for row_col_index, cell_dict in self.cell_data.items():
            keys = tuple((int(i) for i in row_col_index.split(':')))
            if XBRLTableTagEnum.is_bold.value not in cell_dict:
                cell_dict[XBRLTableTagEnum.is_bold.value] = False
            if XBRLTableTagEnum.left_indentation.value not in cell_dict:
                cell_dict[XBRLTableTagEnum.left_indentation.value] = None
            self._cells[keys] = (
                self._populate_cell(keys=keys, cell_dict=cell_dict))

    @staticmethod
    def _populate_cell(keys: tuple, cell_dict: dict) -> 'HTMLCell':
        """ Creates `TDTag` object for each item in Table"""
        return HTMLCell(
            html_id=cell_dict[XBRLCellEnum.id.value],
            xpath=cell_dict[XBRLCellEnum.xpath.value],
            row_index=keys[1],
            col_index=keys[0],
            col_span=cell_dict[XBRLCellEnum.col_span.value],
            row_span=cell_dict[XBRLCellEnum.row_span.value],
            raw_value=cell_dict[XBRLCellEnum.raw_value.value],
            is_bold=cell_dict[XBRLCellEnum.is_bold.value],
            left_indentation=cell_dict[XBRLCellEnum.left_indentation.value]
        )

    def to_dict(self) -> dict:
        """Converts HTMLTableTag to dict"""
        table_dict = {
            XBRLTableTagEnum.xpath.value: self.xpath,
            XBRLTableTagEnum.title.value: self.title,
            XBRLTableTagEnum.anchor_id.value: self.html_id,
            XBRLTableTagEnum.cells.value: self.cell_data,
        }
        output_dict = {
            XBRLTableTagEnum._id.value: self.tag_id,
            XBRLTableTagEnum.value.value: '',
            XBRLTableTagEnum.ocr_value.value: '',
            XBRLTableTagEnum.is_table.value: True,
            XBRLTableTagEnum.table.value: table_dict,
            XBRLTableTagEnum.source.value: self.source_ids}

        return output_dict

    @classmethod
    def construct_from_raw(cls, raw: dict, html: HTML) -> 'HTMLTableTag':
        """Builds HTMLTableTag objeTct from pycognaize raw data
        :param raw: pycognaize field's tag info
        :param html: HTML

        :return:
        """
        tag_id = raw[XBRLTableTagEnum._id.value]
        ocr_value = raw[XBRLTableTagEnum.ocr_value.value]
        value = raw[XBRLTableTagEnum.value.value]
        table_raw_data = raw[XBRLTableTagEnum.table.value]
        xpath = table_raw_data[XBRLTableTagEnum.xpath.value]
        title = table_raw_data[XBRLTableTagEnum.title.value]
        html_id = table_raw_data[XBRLTableTagEnum.anchor_id.value]
        cell_data = table_raw_data[XBRLTableTagEnum.cells.value]
        source_ids = raw[XBRLTableTagEnum.source.value]
        return cls(tag_id=tag_id, ocr_value=ocr_value, value=value,
                   is_table=True, xpath=xpath, title=title,
                   html_id=html_id, cell_data=cell_data,
                   html=html, source_ids=source_ids)

    def _build_df(self) -> pd.DataFrame:
        """Build pandas data frame using `TDTag` Cells

        :return: DataFrame object,where each cell contains an TDTag
            object with the html_id and values from the annotated
            document
        """
        cols = set()
        rows = set()
        for cell_ in self.cells.values():
            cols.add(cell_.col_index - 1)
            rows.add(cell_.row_index - 1)
        cols = list(cols)
        rows = list(rows)
        cols.sort()
        rows.sort()
        df = pd.DataFrame(columns=cols, index=rows)
        for cell_ in self.cells.values():
            row_index = cell_.row_index - 1
            col_index = cell_.col_index - 1
            for col_n in range(col_index, col_index + cell_.col_span):
                for row_n in range(row_index, row_index + cell_.row_span):
                    df[col_n][row_n] = TDTag(is_table=False,
                                             html_id=cell_.html_id,
                                             xpath=cell_.xpath,
                                             raw_value=cell_.raw_value,
                                             raw_ocr_value=cell_.raw_value,
                                             field_id='',
                                             tag_id=self.tag_id,
                                             row_index=row_index,
                                             col_index=col_index)
        return df


class HTMLCell:
    """Represents cell tag for XBRL tables"""
    def __init__(self, row_index: int, col_index: int,
                 col_span: int, row_span: int,
                 html_id: List[str], xpath: str,
                 raw_value: str, is_bold: False,
                 left_indentation: None):
        self._row_index = row_index
        self._col_index = col_index
        self._col_span = col_span
        self._row_span = row_span
        self._html_id = html_id
        self._xpath = xpath
        self._raw_value = raw_value
        self._is_bold = is_bold
        self._left_indentation = left_indentation

    @property
    def row_index(self) -> int:
        return self._row_index

    @property
    def col_index(self) -> int:
        return self._col_index

    @property
    def col_span(self) -> int:
        return self._col_span

    @property
    def row_span(self) -> int:
        return self._row_span

    @property
    def html_id(self):
        return self._html_id

    @property
    def xpath(self):
        return self._xpath

    @property
    def raw_value(self) -> str:
        return self._raw_value

    @property
    def is_bold(self) -> bool:
        return self._is_bold

    @property
    def left_indentation(self) -> str:
        return self._left_indentation

    @classmethod
    def construct_from_raw(cls, raw: dict) -> 'HTMLCell':
        """Build TDTag from pycognaize raw data

        :param raw: pycognaize field's tag info
        :return:
        """
        source_data = raw[XBRLCellEnum.source.value]
        row_index = source_data[XBRLCellEnum.row_index.value]
        col_index = source_data[XBRLCellEnum.col_index.value]
        col_span = source_data[XBRLCellEnum.col_span.value]
        row_span = source_data[XBRLCellEnum.row_span.value]
        raw_value = raw[XBRLCellEnum.raw_value.value]
        html_id = source_data[XBRLCellEnum.html_id.value]
        xpath = source_data[XBRLCellEnum.xpath.value]
        is_bold = (source_data[XBRLCellEnum.is_bold.value]
                   if XBRLCellEnum.is_bold.value in source_data else False)
        left_indentation = (source_data[XBRLCellEnum.left_indentation.value]
                            if XBRLCellEnum.left_indentation.value else None)
        return cls(html_id=html_id, xpath=xpath, row_index=row_index,
                   col_index=col_index, col_span=col_span, row_span=row_span,
                   raw_value=raw_value, is_bold=is_bold,
                   left_indentation=left_indentation)

    def to_dict(self) -> dict:
        """Converts cell to dict"""
        cell_dict = {
            XBRLCellEnum.col_span.value: self.col_span,
            XBRLCellEnum.row_span.value: self.row_span,
            XBRLCellEnum.html_id.value: self.html_id,
            XBRLCellEnum.xpath.value: self.xpath,
            XBRLCellEnum.raw_value.value: self.raw_value,
            XBRLCellEnum.left_indentation.value: self.left_indentation,
            XBRLCellEnum.is_bold.value: self.is_bold}
        return {f"{self.col_index}:{self.row_index}": cell_dict}


class TDTag(HTMLTag):
    def __init__(self, raw_value: str, raw_ocr_value: str,
                 is_table: bool, html_id: List[str],
                 field_id: Optional[str], tag_id: Optional[str],
                 row_index: int, col_index: int, xpath: str):
        super().__init__(html_id=html_id, xpath=xpath, tag_id=tag_id)
        self._raw_value = raw_value
        self._raw_ocr_value = raw_ocr_value
        self._is_table = is_table
        self._field_id = field_id
        self._row_index = row_index
        self._col_index = col_index

    @property
    def raw_value(self):
        """returns adjusted value"""
        return self._raw_value

    @property
    def raw_ocr_value(self):
        return self._raw_ocr_value

    @property
    def is_table(self):
        return self._is_table

    @property
    def field_id(self):
        return self._field_id

    @property
    def row_index(self):
        return self._row_index

    @property
    def col_index(self):
        return self._col_index

    @classmethod
    def construct_from_raw(cls, raw: dict, html: HTML) -> 'TDTag':
        """Build TDTag from pycognaize raw data
        :param html: HTML
        :param raw: pycognaize field's tag info
        :return:
        """
        source_data = raw[XBRLTagEnum.source.value]
        raw_value = raw[XBRLTagEnum.value.value]
        raw_ocr_value = raw[XBRLTagEnum.ocr_value.value]
        is_table = raw[XBRLTagEnum.is_table.value]
        html_id = source_data[XBRLTagEnum.ids.value]
        field_id = source_data[IqRecipeEnum.field_id.value]
        tag_id = source_data[XBRLTagEnum.tag_id.value]
        row_index = source_data[XBRLTagEnum.row_index.value]
        col_index = source_data[XBRLTagEnum.col_index.value] - 1
        xpath = source_data[XBRLTagEnum.xpath.value]
        return cls(html_id=html_id,  raw_value=raw_value,
                   raw_ocr_value=raw_ocr_value, is_table=is_table,
                   field_id=field_id, tag_id=tag_id,
                   row_index=row_index, col_index=col_index,
                   xpath=xpath)

    def to_dict(self) -> dict:
        """Converts tag to dict"""
        tag_info = {
            XBRLTagEnum.ids.value: self.html_id,
            IqRecipeEnum.field_id.value: self.field_id,
            XBRLTagEnum.tag_id.value: self.tag_id,
            XBRLTagEnum.row_index.value: self.row_index,
            XBRLTagEnum.col_index.value: self.col_index + 1,
            XBRLTagEnum.xpath.value: self.xpath,
        }
        output_dict = {
            ID: str(bson.ObjectId()),
            XBRLTagEnum.value.value: self.raw_value,
            XBRLTagEnum.ocr_value.value: self.raw_ocr_value,
            XBRLTagEnum.is_table.value: self.is_table,
            XBRLTagEnum.source.value: tag_info,
         }
        return output_dict



import logging
import os

from bs4 import BeautifulSoup

from pycognaize.login import Login
from pycognaize.common.enums import StorageEnum
from pycognaize.common.utils import cloud_interface_login


class HTML:
    """Represents html of a xbrl document in pycognaize"""
    def __init__(self, path: str, document_id: str) -> None:
        """
        :param path: Local or remote path to the document folder,
            which includes the html file
        """
        self._login_instance = Login()
        self.ci = cloud_interface_login(self._login_instance)
        self._path = self._validate_path(path, document_id)
        self._html_file = None
        self._html_soup = None

    @property
    def path(self) -> str:
        """Path of the source document"""
        return self._path

    @property
    def html_soup(self):
        if self._html_soup is None:
            if self.path:
                self._html_soup = BeautifulSoup(self._get_html(),
                                                features="html.parser")
        return self._html_soup

    def _validate_path(self, path: str, document_id: str) -> str:
        """
        If the input path contains a directory with the same name
            as the document ID, the valid path
            is the combination of the input path and the document id
        If the input path contains a file named `source.html`,
            the valid path is simply the input path.
        Otherwise, the function returns an empty string as the valid path.

        :param path: path of the source document
        :param document_id: document id
        :return: valid path for the `source.html` file
            corresponding to the document id
        """
        valid_path = ''
        if self.ci.isdir(os.path.join(path, document_id)):
            valid_path = os.path.join(path, document_id)
        elif StorageEnum.html_file.value in self.ci.listdir(path):
            valid_path = path
        return valid_path

    def _read_html(self, path: str) -> str:
        if self._html_file is None:
            with self.ci.open(path, 'r') as file:
                self._html_file = file.read()
        return self._html_file

    def _get_html(self):
        html_bytes = None
        uri = os.path.join(self.path, StorageEnum.html_file.value)
        try:
            html_bytes = self._read_html(path=uri)
        except FileNotFoundError as e:
            logging.debug(
                f"Unable to get the html: {e}")
        return html_bytes



import codecs
import os.path
import re

from setuptools import setup


here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    return codecs.open(os.path.join(here, *parts), 'r').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup_options = dict(
    license="Apache License 2.0",
    python_requires=">= 3.7",
    entry_points='''
        [console_scripts]
        cognaize=bin.cognaize:main''',
    project_urls={
        'Source': 'https://github.com/cognaize/pycognaize',
        'Reference': 'https://pycognaize.readthedocs.io/en/latest/',
        'Changelog': 'https://github.com/cognaize/pycognaize/blob/master/CHANGELOG.md',
    },
)

setup_options['console'] = ['bin/cognaize']


setup(**setup_options)



from collections import defaultdict
from collections.abc import Mapping


class LazyGroupDict(Mapping):
    """Contains groups included in the document"""

    def __init__(self, document_fields: dict):
        self._document_fields = document_fields
        self._groups = None
        self._groups_by_key = None

    @property
    def groups_by_key(self) -> dict:
        if self._groups_by_key is None:
            self.__create_groups_by_key(self._document_fields)
        return self._groups_by_key

    @property
    def groups(self) -> dict:
        if self._groups is None:
            self.__create_groups(self._document_fields)
        return self._groups

    def __create_groups(self, document_fields: dict):
        """Add fields to group """
        groups = defaultdict(dict)
        group_names = set()

        for field_name in document_fields.keys():
            group_names.update(
                [item.group_name for item in document_fields[field_name]
                 if item.group_name])

        for group_name in group_names:
            groups[group_name] = defaultdict(list)
            for field_name in document_fields.keys():
                for item in document_fields[field_name]:
                    if item.group_name == group_name:
                        groups[group_name][item.group_key].append(item)

        self._groups = groups

    def __create_groups_by_key(self, document_fields: dict):
        """Add fields to group """
        groups_by_key = defaultdict(list)
        for field_name in document_fields.keys():
            for item in document_fields[field_name]:
                groups_by_key[item.group_key].append((field_name, item))

        self._groups_by_key = groups_by_key

    def __getitem__(self, group_name) -> dict:
        """The Document object, retrieved from provided path
        Note: Path can be both local and remote
        """
        if group_name in self.groups:
            return self.groups[group_name]
        else:
            raise KeyError

    def __iter__(self):
        ...

    def __len__(self):
        ...



from pycognaize.common.utils import convert_coord_to_num
from pycognaize.document.tag.tag import BoxTag
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pycognaize.document.page import Page

from pycognaize.common.decorators import module_not_found


class SpanTag(BoxTag):
    """Represents a Tag included in a spanfield"""

    def __init__(self, left, right, top, bottom,
                 page: 'Page',
                 raw_value: str, raw_ocr_value: str):
        super().__init__(left=left, right=right, top=top, bottom=bottom,
                         page=page)
        self.raw_value = raw_value
        self.raw_ocr_value = raw_ocr_value
        self._spacy_doc = None

    @classmethod
    def construct_from_raw(cls, raw: dict, page: 'Page') -> 'SpanTag':
        """Create a SpanTag from a raw dictionary"""
        left = convert_coord_to_num(raw['left'])
        top = convert_coord_to_num(raw['top'])
        height = convert_coord_to_num(raw['height'])
        width = convert_coord_to_num(raw['width'])
        right = left + width
        bottom = top + height
        raw_value = raw['value']
        raw_ocr_value = raw['ocrValue']
        tag = cls(left=left, right=right, top=top, bottom=bottom,
                  page=page, raw_value=raw_value, raw_ocr_value=raw_ocr_value)
        return tag

    @module_not_found
    def __create_spacy_doc(self):
        """Creates spacy nlp object from raw value"""
        import spacy
        nlp = spacy.blank("en")
        self._spacy_doc = nlp(self.raw_value)

    def __getitem__(self, val):
        """Returns slice of the span object"""
        return SpanTag(self._left, self._right, self._top,
                       self._bottom, self._page, self.raw_value[val],
                       self.raw_ocr_value[val])



import logging
from itertools import groupby
from typing import Optional

from pycognaize.document.tag.html_tag import HTMLTableTag, TDTag


def filter_out_invalid_tables(tables):
    """
    Filters tables that have tag.
    """
    valid_tables = []
    for table in tables:
        if not table.tags:
            logging.warning('removing table with no tags')
            continue
        valid_tables.append(table)
    return valid_tables


def _sort_table_horizontally(tables, threshold: float):
    """
    Given tables are sorted first horizontally, then vertically.
    """
    groups = []
    for table in tables:
        if not groups:
            groups.append([table])
            continue
        latest_group = groups[-1]
        group_left = min([table.tags[0].left for table in latest_group])
        group_right = max([table.tags[0].right for table in latest_group])
        table_tag = table.tags[0]
        if group_left < table_tag.right and group_right > table_tag.left:
            iou = (
                    (min((table_tag.right, group_right)) -
                     max((table_tag.left, group_left))) /
                    (min(table_tag.right - table.tags[0].left,
                         group_right - group_left)))
            if iou > threshold:
                latest_group.append(table)
                continue
        groups.append([table])

    sorted_tables = []
    for group in groups:
        for table in sorted(group, key=lambda x: x.tags[0].top):
            sorted_tables.append(table)
    return sorted_tables


def assign_indices_to_tables(tables, all_tables: Optional[list] = None,
                             threshold: float = 0.4) -> dict:
    """
    If the document is an XBRL document,
        the function matches the tables based on the ordering of all tables.
    If it's not an XBRL document,
        the tables are grouped by pages and for each page,
        the tables are left sorted and ordered horizontally and vertically.

    Return dict where the keys are indices based above-mentioned ordering
        and the values are the corresponding tables.

    :param tables: a list of tables that need to be indexed
    :param all_tables: a list of all tables in the document.
        This parameter is required if the tables are from an XBRL document
    :param threshold: intersection threshold
    """
    tables_dict = {}
    valid_tables = filter_out_invalid_tables(tables)
    if not valid_tables:
        return tables_dict
    if all(isinstance(table.tags[0], HTMLTableTag) or
           isinstance(table.tags[0], TDTag) for table in valid_tables):
        if not all_tables:
            logging.error('Missing argument: list of all table fields')
            return tables_dict
        all_valid_tables = filter_out_invalid_tables(all_tables)
        tables_html_id_idx_mapping = {
            table.tags[0].html_id: (idx, 0)
            for idx, table in enumerate(all_valid_tables, start=1)}
        tables_dict = {
            tables_html_id_idx_mapping[table.tags[0].html_id]: table
            for table in valid_tables}
    else:
        sorted_tables = sorted(valid_tables,
                               key=lambda x: x.tags[0].page.page_number)
        grouped_tables = {page: list(table) for page, table in
                          groupby(sorted_tables,
                                  key=lambda x: x.tags[0].page.page_number)}
        for page, page_tables in grouped_tables.items():
            sorted_page_tables = sorted(page_tables,
                                        key=lambda x: x.tags[0].left)
            final_ordered_tables = _sort_table_horizontally(
                sorted_page_tables, threshold=threshold)
            tables_dict.update({(page, idx): table
                                for idx, table in
                                enumerate(final_ordered_tables)})
    return tables_dict



import math
from typing import Union


class Confidence:
    """
    This class is used by tag object to store confidence values
    of possible classes.
    """

    def __init__(self, confidences=None):
        """
        Initialize class with confidence scores.
        :param confidences: dictionary of class names and their confidence
        """
        self._is_finalized = False
        if confidences is not None and isinstance(confidences, dict):
            self.confidences = confidences
            self.finalize()
        else:
            self.confidences = {}

    def add_class(self, class_name: str,
                  confidence: Union[float, int]) -> None:
        """Add new class for confidence scores
        :param class_name: name of the class
        :param confidence: confidence score for the class
        """
        self.confidences[class_name] = confidence

    def number_of_classes(self) -> int:
        """Returns number of classes"""
        return len(self.confidences)

    def class_names(self) -> list:
        """Returns class names"""
        return list(self.confidences.keys())

    def get_confidence(self) -> dict:
        """Returns confidence scores if finalized"""
        if not self._is_finalized:
            raise ValueError('Confidence scores have to be finalized.')
        return self.confidences

    @property
    def is_finalized(self) -> bool:
        return self._is_finalized

    def finalize(self, tolerance=0.0001) -> None:
        """Finalize confidence scores by checking if all scores sum to 1
        :param tolerance: tolerance for comparing sum of confidences (float)
        with 1
        """
        sum_scores = sum(self.confidences.values())
        if not math.isclose(sum_scores, 1, rel_tol=tolerance) and \
                self.number_of_classes() > 0:
            raise ValueError('Confidence scores have to sum up to 1.')
        self._is_finalized = True

    def __getitem__(self, item: str) -> Union[float, int]:
        if not self._is_finalized:
            raise ValueError('Confidence scores have to be finalized.')
        if item in self.confidences:
            return self.confidences[item]
        else:
            raise KeyError(f'Class {item} not found in confidence scores.')

    def __setitem__(self, key, value) -> None:
        self.confidences[key] = value



import os
import logging
from typing import Mapping, Tuple

from pycognaize.login import Login
from pycognaize.common import utils
from pycognaize.common.enums import EnvConfigEnum
from pycognaize.common.exceptions import AuthenthicationError
from pycognaize.common.lazy_dict import LazyDocumentDict


class Snapshot:
    """A snapshot of annotated documents from one or more collections"""

    def __init__(self, path: str):
        self._path = path
        self._documents = LazyDocumentDict(doc_path=path, data_path=path)

    @property
    def documents(self) -> Mapping:
        """Mapping of document ids to documents
        :return Mapping:  LazyDocumentDict Object
        """
        return self._documents

    @classmethod
    def get_by_id(cls, snapshot_id: str) -> 'Snapshot':
        """Returns the Snapshot Object"""
        login_instance = Login()
        if login_instance.logged_in:
            snapshot_dir = login_instance.snapshot_root
            snapshot_path = os.path.join(snapshot_dir, snapshot_id)
            return cls(path=snapshot_path)
        else:
            snapshot_dir = os.environ[EnvConfigEnum.SNAPSHOT_PATH.value]
            snapshot_id = os.environ[EnvConfigEnum.SNAPSHOT_ID.value]
            snapshot_path = os.path.join(snapshot_dir, snapshot_id)
            return cls(path=snapshot_path)

    @classmethod
    def download(cls, snapshot_id: str, destination_dir: str) -> \
            Tuple['Snapshot', str]:
        """Downloads snapshot to specified destination"""
        login_instance = Login()

        if login_instance.logged_in:
            snapshot_path = os.path.join(login_instance.snapshot_root,
                                         snapshot_id)
            ci = utils.cloud_interface_login(login_instance)
            ci.copy_dir(snapshot_path, destination_dir)

            logging.info(f"Snapshot {snapshot_id} downloaded to "
                         f"{destination_dir}. To use the snapshot, check our "
                         f"documentation at: ",
                         "http://pycognaize-docs.com."
                         "s3-website.us-east-2.amazonaws.com")

            return cls(path=snapshot_path), \
                os.path.join(destination_dir, snapshot_id),
        else:
            raise AuthenthicationError()

    @classmethod
    def _snapshot_path(cls) -> str:
        """Identify and return the snapshot path"""
        login_instance = Login()
        if login_instance.logged_in:
            snapshot_dir = login_instance.snapshot_root
            snapshot_id = os.environ[EnvConfigEnum.SNAPSHOT_ID.value]
            snapshot_path = os.path.join(snapshot_dir, snapshot_id)
        else:
            snapshot_dir = os.environ[EnvConfigEnum.SNAPSHOT_PATH.value]
            snapshot_id = os.environ[EnvConfigEnum.SNAPSHOT_ID.value]
            snapshot_path = os.path.join(snapshot_dir, snapshot_id)
        return snapshot_path

    @classmethod
    def get(cls) -> 'Snapshot':
        """Read the snapshot object from local storage and return it"""
        return cls(path=cls._snapshot_path())
